/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * LISTFUNC_H
 *
 * PROGRAMMER : Gregg A. Tavares 
 *    VERSION : 01.001
 *    CREATED : 08/04/89 
 *   MODIFIED : 10/02/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *    Functions and definitions for dealing with linked lists.
 *
 * HISTORY
 *		08/07/89 (GAT) - Added Macros StartOfList() and EndOfList.
 *		09/07/89 (GAT) - Fixed IsEmpty Macro.
 *		01/09/90 (GAT) - Changed type to be compatible with AmigaOS.
 *		05/25/91 (RGM) - Added include for <stddef.h>.
 *		10/05/91 (RGM) - Added #if for SAS/Lattice Amiga includes.
 *		10/30/91 (RGM) - Added	#define EndOfList	IsEOList and 
 *								#define StartOfList	IsSOList.
 *
*/

#ifndef ECHIDNA_LISTFUNC_H
#define ECHIDNA_LISTFUNC_H

#include <echidna/platform.h>
#include "switches.h"

#ifdef __cplusplus
extern "C" {
#endif

#include <string.h>
#include <echidna/etypes.h>
#if __NOT_ANSI__
#else
#include <stdlib.h>
#include <stddef.h>
#endif

#if LATTICE
#include <exec/types.h>
#include <proto/exec.h>
#endif

/********************************** TYPES *********************************/

/*
** NodeType
**
** Structure for tracking nodes in a doubly-linked list.
**
** A good way to use this structure for your own lists is to make it the
** first field of the structure you need in a list.  ie.
**
**      typedef struct {
**          NodeType       Node;       ** Node structure for list routines ** 
**          int            MyData1;    ** Your fields...                   **
**          long           MyData2;
**          void          *MyData3;
**      } MyType;
**
** If your structures follow this example the list routines will have no
** problem handling them.
**
** By definition, the first valid element of a list is the 'Head' node and 
** the last valid element is the 'Tail' node.  
**
** There are 4 Macros to help in dealing with nodes in a list.  All of them
** take as argument a pointer to a NodeType.  
**
** The macro Next(node) returns a pointer to the node in the list which 
** follows the argument node.
** The macro Prev(node) returns a pointer to the node in the list which 
** precedes the argument node.
**
** There are two "pseudo-nodes" in every list.  One appears before the Head
** node and the other appears after the Tail node.  These nodes tell you
** when to stop traversing the list.  
**
** After using node=Next(node), the macro IsEOList(node) will tell you if 
** you're pointing at the end pseudo-node.
** After using node=Prev(node), the macro IsSOList(node) will tell you if
** you're pointing at the start pseudo-node.
**
** Here is an example of traversing a list from head to tail.
**
**    ListType *MyList;
**    NodeType *nd;
**    ...
**    nd = Head(MyList);	
**    while (!IsEOList(nd)) {
**       ...
**       nd = Next(nd);
**    }
** 
** Here is an example of traversing a list from tail to head.
**
**    ListType *MyList;
**    NodeType *nd;
**    ...
**    nd = Tail(MyList); 
**    while (!IsSOList(nd)) {
**       ...
**       nd = Prev(nd);
**    }
**
** SEE ALSO
**   Everything else in this file!!!
*/

typedef struct MinNodeType {
	struct	NodeType	*Next;
	struct	NodeType	*Prev;
} MinNodeType;

typedef struct MinNode {
	struct	NodeType	*Next;
	struct	NodeType	*Prev;
} MinNode;

typedef struct NodeType {
	struct	NodeType	*Next;
	struct	NodeType	*Prev;
	UBYTE				 Type;
	BYTE				 Priority;
	char				*Name;
} NodeType;

/*
** ListType
**
** Structure to keep track of head and tail of a list.
**
** You always need one of these before you can have a list.  CreateList()
** will create one for you or you can create them statically and use
** InitList() to set them up.
**
** There are three Macros that may be used on a list.  All three require a
** pointer to a ListType structure.  Head(list) will return a pointer
** to the first node in the list.  Tail(list) will return a pointer to the
** last node in the list.  IsEmpty(list) is TRUE if there are no elements
** in the list.
*/

typedef struct {
	NodeType	*Head;
	NodeType	*Tail;
	NodeType	*TailPred;
	UBYTE		 Type;
	BYTE		 Priority;
	char	 	*Name;
} ListType;

typedef void *(*OpOnListFunc)(void *);

/********************************* MACROS *********************************/

#define EndOfList	IsEOList
#define StartOfList	IsSOList

#if !DEBUG_LISTFUNC
#define Next(node)		((void *)(((NodeType *)(node))->Next))
#define Prev(node)		((void *)(((NodeType *)(node))->Prev))
#define IsEOList(node)	(((NodeType *)(node))->Next == NULL)
#define IsSOList(node)	(((NodeType *)(node))->Prev == NULL)
#else
#define CreateNode		CreateNode2
#define IsEOList(node)	(Next(node) == NULL)
#define IsSOList(node)	(Prev(node) == NULL)
#endif

#define Head(list)		((void *)(((ListType *)(list))->Head))
#define Tail(list)		((void *)(((ListType *)(list))->TailPred))
#define IsEmpty(list)	(((ListType *)(list))->TailPred == ((NodeType *)(list)))
#define FindIName(node,name)	Find_A_Name((void *)(node),(name), 1)
#define NodeName(nd)	(((NodeType *)(nd))->Name)

#if __AMIGAOS__
#if AZTEC_C
#include <functions.h>
#endif
#define clearstruct(data,size)	memset((char *)(data),0,(size_t)(size))
#define	Remove(node)			Remove((struct Node *)(node))
#define AddHead(list,node)		AddHead((struct List *)(list),(struct Node *)(node))
#define AddTail(list,node)		AddTail((struct List *)(list),(struct Node *)(node))
#define RemHead(list)			((void *)(RemHead((struct List *)(list))))
#define RemTail(list)			((void *)(RemTail((struct List *)(list))))
#define FindName(list,name)		((void *)(FindName((struct List *)(list),(name))))
#endif

#if __MSDOS__
#define clearstruct(data,size)	memset((char *)(data),0,(size_t)(size))
#define FindName(node,name)		Find_A_Name((void *)(node),(name), 0)
#endif

#if __MACOS__
#define	clearstruct(data,size)	memset((char *)(data),0,(size_t)(size))
#endif

#if __EGGS_SGI__
#define	clearstruct(data,size)	memset((char *)(data),0,(size_t)(size))
#endif

//
// new style library names
//
#define LST_Next			Next
#define LST_Prev			Prev
#define LST_IsEOList		IsEOList
#define LST_IsSOList		IsSOList
#define LST_EndOfList		EndOfList
#define LST_StartOfList		StartOfList
#define LST_Head			Head
#define LST_Tail			Tail
#define LST_IsEmpty			IsEmpty
#define LST_FindIName		FindIName
#define LST_NodeName		NodeName
#define LST_CreateNode		CreateNode
#define LST_AddName			AddName
#define LST_FreeName		FreeName
#define LST_DeleteNode		DeleteNode
#define LST_InitList		InitList
#define LST_CreateList		CreateList
#define LST_EmptyList		EmptyList
#define LST_DeleteList		DeleteList
#define LST_InsertBefore	InsertBefore
#define LST_InsertAfter		InsertAfter
#define LST_Find_A_Name		Find_A_Name
#define LST_OpOnList		OpOnList
#define LST_InsertList		InsertList
#define LST_MergeLists		MergeLists
#define LST_ConcatLists		ConcatLists
#define LST_InsertInOrder	InsertInOrder
#define	LST_Remove			Remove
#define LST_AddHead			AddHead
#define LST_RemHead			RemHead
#define LST_AddTail			AddTail
#define LST_RemTail			RemTail
#define LST_FindName		FindName

/********************************* GLOBALS ********************************/

/*************************** FUNCTION PROTOTYPES **************************/

#if __NOT_ANSI__

extern char		*AddName ();
extern void		 FreeName ();
extern void		*CreateNode ();
extern void		*DeleteNode ();
extern void		 InitList ();
extern ListType	*CreateList ();
extern void		 EmptyList ();
extern ListType	*DeleteList ();
extern void		 InsertBefore ();
extern void		 InsertAfter ();
extern void		*InsertInOrder ();

#if !__AMIGAOS__
extern void		 Remove ();
extern void		 AddHead ();
extern void		*RemHead ();
extern void		 AddTail ();
extern void		*RemTail ();
extern void		*FindName ();
#endif

#else

#if DEBUG_LISTFUNC
extern void		*Next (void *);
extern void		*Prev (void *);
extern void		*CreateNode2 (size_t size, char *name);
#endif
extern void		*CreateNode (size_t size, char *name);

extern char		*AddName (void *node, char *name);
extern void		 FreeName (void *node);
extern void		*DeleteNode (void *node);
extern void		 InitList (ListType *list);
extern ListType	*CreateList (char *name);
extern void		 EmptyList (ListType *list);
extern ListType	*DeleteList (ListType *list);
extern void		 InsertBefore (void *oldnode, void *newnode);
extern void		 InsertAfter (void *oldnode, void *newdnode);
extern void		*Find_A_Name (void *list, char *name, short insensitive);
extern long		 OpOnList (void *node, OpOnListFunc func);
extern void		*InsertList (void *node1, void *node2);
extern void		*MergeLists (void *node1, void *node2);
extern ListType	*ConcatLists (ListType *list1, ListType *list2);
extern void		*InsertInOrder (void *inode, void *node);

#if !__AMIGAOS__
extern void		 Remove (void *node);
extern void		 AddHead (ListType *list, void *node);
extern void		*RemHead (ListType *list);
extern void		 AddTail (ListType *list, void *node);
extern void		*RemTail (ListType *list);
#endif

#endif

#ifdef __cplusplus
}
#endif

#endif /* ECHIDNA_LISTFUNC_H */

#if 0

/**************************************************************************
 *
 * clearstruct
 *
 * SYNOPSIS
 *   void clearstruct (char *data, UWORD size)
 *
 * FUNCTION
 *   Clears structure pointed to my 'data' of size 'size' to all zeros.
 *
 * INPUTS
 *   data = pointer to structure to be cleared.
 *   size = size of structure to be cleared.
 *
 * RESULTS
 *   strcuture is cleared.
 *
 * EXAMPLE
 *   clearstruct (mylist, sizeof(*mylist));
 *   clearstruct (mylist, sizeof(ListType));
*/
/**************************************************************************
 *
 * AddName
 *
 * SYNOPSIS
 *   char *AddName (NodeType *node, char *name);
 *
 * FUNCTION
 *   Adds a name to a node.  This function first allocates memory the size
 *   of the string pointed to by 'name' and then attaches that memory to
 *   the 'Name' field of 'node'.
 *
 * INPUTS
 *   node = pointer to a NodeType or any structure whose first field is a
 *          NodeType.
 *   name = pointer to a NULL terminated string.
 *
 * RESULTS
 *   returns NULL if an error occured (out of memory)
 *
 * SEE ALSO
 *   FreeName, CreateNode, DeleteNode.
*/
/**************************************************************************
 *
 * FreeName
 *
 * SYNOPSIS
 *   void FreeName (NodeType *node);
 *
 * FUNCTION
 *   Frees any name associated with 'node'.  node->Name must point to a
 *   Null terminated string (that was allocated dynamically not statically)
 *   node->Name may be NULL.
 *
 * INPUTS
 *   node = pointer to a NodeType or any structure whose first field is a
 *          NodeType.
 *
 * RESULTS
 *   memory for string that was attached to node is freed.
 *
 * SEE ALSO
 *    AddName, DeleteNode,
*/
/**************************************************************************
 *
 * CreateNode
 *
 * SYNOPSIS
 *   void *CreateNode (int size, char *name);
 *
 * FUNCTION
 *   Allocates memory for and returns a pointer to that memory for a
 *   structure who's first field is assumed to be a NodeType.  If the
 *   'name' parameter is not NULL then AddName() is called for the name
 *   to be duplicated and attached to the new node.
 *
 * INPUTS
 *   size = size of structure to be allocated (ie. sizeof (MyType))
 *   char = pointer to a NULL terminated string or just NULL.
 *
 * RESULTS
 *   pointer to structure allocated is returned or NULL if error
 *   (out of memory)
 *
 * SEE ALSO
 *   DeleteNode
*/
/**************************************************************************
 *
 * DeleteNode
 *
 * SYNOPSIS
 *   void *DeleteNode (NodeType *node)
 *
 * FUNCTION
 *   Deletes a node (frees it from memory) and any attached name string
 *   is also freed.  It is assumed 'node' is not part of a list.
 *
 * INPUTS
 *   node = pointer to a NodeType or any structure whose first field is a
 *          NodeType.
 *
 * RESULTS
 *    always returns NULL.
 *
 * SEE ALSO
 *   CreateNode, DeleteList, EmptyList
*/
/**************************************************************************
 *
 * InitList
 *
 * SYNOPSIS
 *   void InitList (ListType *list);
 *
 * FUNCTION
 *   Initializes a ListType structure to an empty list.  If you create your
 *   own ListType structures (ie. without calling CreateList()) then you
 *   must call InitList on each one before you can use the lists with the
 *   list routines.
 *
 * INPUTS
 *   list = pointer to a ListType.
 *
 * SEE ALSO
 *   CreateList
*/
/**************************************************************************
 *
 * CreateList
 *
 * SYNOPSIS
 *   ListType *CreateList (char *name)
 *
 * FUNCTION
 *   This function allocates a ListType structure.  Initializes it to an
 *   empty list (by calling InitList()) and if the 'name' parameter is
 *   not NULL then the string pointed to by 'name' is duplicated and
 *   attached to the list.
 *
 * INPUTS
 *   name = pointer to a NULL terminated string or just NULL.
 *
 * RESULTS
 *   pointer to newly created ListType structure or NULL if error
 *   (out of memory)
 *
 * SEE ALSO
 *   DeleteList, EmptyList
*/
/**************************************************************************
 *
 * EmptyList
 *
 * SYNOPSIS
 *   void EmptyList (ListType *list)
 *
 * FUNCTION
 *   This function frees from memory every node still attached to this
 *   list and every string of every node attached to this list. You should
 *	 not call this function if you haven't freed anything pointed to by
 *   nodes of this list.
 *
 * INPUTS
 *   list = pointer to a ListType structure
 *
 * RESULTS
 *
 * SEE ALSO
 *   DeleteNode, CreateList
*/
/**************************************************************************
 *
 * DeleteList
 *
 * SYNOPSIS
 *   void *DeleteList (ListType *list)
 *
 * FUNCTION
 *   This function frees from memory every node still attached to this
 *   list, every string of every node attached to this list, Any string
 *   attached directly to this list, and the ListType structure itsself.
 *   You should not call this function if you created your ListType
 *   structure statically or if you haven't freed anything pointed to by
 *   nodes of this list.
 *
 * INPUTS
 *   list = pointer to a ListType structure
 *
 * RESULTS
 *   always returns NULL.
 *
 * SEE ALSO
 *   DeleteNode, CreateList, EmptyList
*/
/**************************************************************************
 *
 * InsertBefore
 *
 * SYNOPSIS
 *   void InsertBefore (NodeType *node1, NodeType *node2)
 *
 * FUNCTION
 *   Inserts 'node2' into list before 'node1'.  'node1' is assumed to be
 *   part of a list.  'node2' is assumed to NOT be part of a list.
 *
 * INPUTS
 *   node1 = pointer to a NodeType or any structure whose first field is a
 *           NodeType.
 *   node2 = pointer to a NodeType or any structure whose first field is a
 *           NodeType.
 *
 * RESULTS
 *   'node2' is now part of list linked as the node that comes before
 *   'node1'.
 *
 * SEE ALSO
 *   AddHead, AddTail, InsertAfter.
*/
/**************************************************************************
 *
 * InsertAfter
 *
 * SYNOPSIS
 *   void InsertAfter (NodeType *node1, NodeType *node2)
 *
 * FUNCTION
 *   Inserts 'node2' into list After 'node1'.  'node1' is assumed to be
 *   part of a list.  'node2' is assumed to NOT be part of a list.
 *
 * INPUTS
 *   node1 = pointer to a NodeType or any structure whose first field is a
 *           NodeType.
 *   node2 = pointer to a NodeType or any structure whose first field is a
 *           NodeType.
 *
 * RESULTS
 *   'node2' is now part of list linked as the node that comes after 'node1'.
 *
 * SEE ALSO
 *   AddHead, AddTail, InsertBefore.
*/
/**************************************************************************
 *
 * Remove
 *
 * SYNOPSIS
 *   void Remove (NodeType *node);
 *
 * FUNCTION
 *   Unlinks 'node' from its list.
 *
 * INPUTS
 *   node = pointer to a NodeType or any structure whose first field is a
 *          NodeType.
 *
 * RESULTS
 *   node is now unlinked from list and can be freed or attached to another
 *   list or ...
 *
 * SEE ALSO
 *   RemTail, RemHead.
*/
/**************************************************************************
 *
 * AddHead
 *
 * SYNOPSIS
 *   void AddHead (ListType *list, NodeType *node)
 *
 * FUNCTION
 *   Adds 'node' to 'list' as the first node in the list.
 *
 * INPUTS
 *   list = pointer to a ListType structure
 *   node = pointer to a NodeType or any structure whose first field is a
 *          NodeType.
 *
 * RESULTS
 *   'node' is now attaced to 'list' as head of list.
 *
 * SEE ALSO
 *   RemHead, AddTail, InsertBefore, InsertAfter.
*/
/**************************************************************************
 *
 * RemHead
 *
 * SYNOPSIS
 *   void *RemHead (ListType *list);
 *
 * FUNCTION
 *   Returns a pointer to first node in 'list' or NULL if none exist.
 *
 * INPUTS
 *   list = Pointer to a ListType structure.
 *
 * RESULTS
 *   if any nodes are in the list the first one is removed from the list
 *   and a pointer to it is returned, otherwise NULL is returned.
 *
 * SEE ALSO
 *   AddHead, RemTail, Remove.
*/
/**************************************************************************
 *
 * AddTail
 *
 * SYNOPSIS
 *   void AddTail (ListType *list, NodeType *node)
 *
 * FUNCTION
 *   Adds 'node' to 'list' as the list node in the list.
 *
 * INPUTS
 *   list = pointer to a ListType structure
 *   node = pointer to a NodeType or any structure whose first field is a
 *          NodeType.
 *
 * RESULTS
 *   'node' is now attaced to 'list' as tail of list.
 *
 * SEE ALSO
 *   RemTail, AddHead, InsertBefore, InsertAfter.
*/
/**************************************************************************
 *
 * RemTail
 *
 * SYNOPSIS
 *   void *RemTail (ListType *list);
 *
 * FUNCTION
 *   Returns a pointer to last node in 'list' or NULL if none exist.
 *
 * INPUTS
 *   list = Pointer to a ListType structure.
 *
 * RESULTS
 *   if any nodes are in the list the last one is removed from the list
 *   and a pointer to it is returned, otherwise NULL is returned.
 *
 * SEE ALSO
 *   AddTail, RemHead, Remove.
*/

#endif

