/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * PHOTOSHP.C
 *
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 09/13/94
 *   MODIFIED : 03/13/95
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oO///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		Load photoshop 2.5 file into a BlockO24BitPixels structure.
 *		This code is designed for 32-bit platforms.
 *
 * HISTORY
 *		09/21/94 (dcc) - compiles cleanly with 32-bit Borland C++ 4.02.
 *		02/05/95 (dcc) - #include "photoshp.h" to verify prototypes.
 *		02/20/95 (GAT) - Don't read directly into data structure as it
 *						 is padded on an SGI.
 *
*/

#include <echidna\platform.h>
#include "switches.h"
#include <echidna\proc.h>

#include <stdio.h>
#include <stdlib.h>

#include <echidna\etypes.h>
#include <echidna\eio.h>
#include <echidna\eerrors.h>
#include <echidna\checkglu.h>
#include <echidna\read24.h>

#include "photoshp.h"					// verify function prototypes

/**************************** C O N S T A N T S ***************************/


/******************************** T Y P E S *******************************/

typedef struct
{
	ULONG	signature;
	UWORD	version;
	UBYTE	reserved[6];
	UWORD	channels;
	ULONG	rows;
	ULONG	columns;
	UWORD	depth;
	UWORD	mode;
}
PSHeader;

/****************************** G L O B A L S *****************************/


/******************************* M A C R O S ******************************/


/***************************** R O U T I N E S ****************************/

typedef struct
{
	UBYTE	*buffer;
	UBYTE	*curPtr;
	long	 size;
	long	 bytesLeft;
}
MEMFILE;

#define	MEMFILE_getc(mf) \
	(((mf)->bytesLeft > 0) ? (((mf)->bytesLeft--),(*mf->curPtr++)) : EOF)

MEMFILE *MEMFILE_Load (char *filename)
{
	static MEMFILE mfX;
	static MEMFILE *mf = &mfX;

	long	len;
	int		fh;

	memset (mf, 0, sizeof (MEMFILE));

	fh  = CHK_ReadOpen (filename);
	len = CHK_FileLength (fh);
	if (len)
	{
		mf->buffer = CHK_AllocateMemory (len, "file buffer");
		if (mf->buffer)
		{
			mf->curPtr    = mf->buffer;
			mf->size      = len;
			mf->bytesLeft = len;

			CHK_ReadOpen (filename);

			CHK_Read (fh, mf->buffer, len);

			CHK_Close (fh);

			return mf;
		}
	}
	return NULL;
}

void MEMFILE_Close (MEMFILE *mf)
{
	CHK_DeallocateMemory (mf->buffer, "file buffer");
}

int MEMFILE_Read(MEMFILE *mf, void *buf, long len)
{
	if (mf->bytesLeft)
	{
		long	copyBytes;

		copyBytes = (mf->bytesLeft < len) ? mf->bytesLeft : len;

		memcpy (buf, mf->curPtr, copyBytes);

		mf->bytesLeft -= copyBytes;
		mf->curPtr    += copyBytes;

		return copyBytes;
	}
	return 0;
}

int MEMFILE_Seek (MEMFILE *mf, long pos, int type)
{
	switch (type)
	{
	case SEEK_SET:
		mf->curPtr    = mf->buffer + pos;
		mf->bytesLeft = mf->size   - pos;
		break;
	case SEEK_CUR:
		mf->curPtr    += pos;
		mf->bytesLeft -= pos;
		break;
	case SEEK_END:
		mf->curPtr     = mf->buffer + mf->size + pos;
		mf->bytesLeft  = -pos;
		break;
	}
	return mf->size - mf->bytesLeft;
}

PROC_LOCAL (BOOL , unpack, (UBYTE *buffer, MEMFILE *mf, long width, long lines, long mod))
{
	char	 i;
	int		 j;
	long	 bytecount;
	long	 d;
	long	 count;
	long	 line;

	for (line = 0; line < lines; line++)
	{
		bytecount = 0;
		while (bytecount < width)
		{
		// read a char & check for eof

			d = MEMFILE_getc(mf);
			if (d == EOF)
			{
				return(1);
			}
			i = (char)d;

			if(i < 0)
			{
			// run data

				count = -i + 1;

				d = MEMFILE_getc(mf);
				if(d == EOF)
				{
					return(1);
				}

				for(j = 0; j < count; j++)
				{
					*buffer = (UBYTE)d;
					buffer += mod;
				}
			}
			else
			{
			// dump data

				count = i + 1;

				for (j = 0; j < count; j++)
				{
					d = MEMFILE_getc(mf);
					if(d == EOF)
					{
						return(1);
					}
					*buffer = (UBYTE)d;
					buffer += mod;
				}
			}

			bytecount += count;
		}
	}
	return 0;
}
PROC_END	// unpack

/*********************************************************************
 *
 * loadPhotoshop
 *
 * SYNOPSIS
 *		void  loadPhotoshop (BlockO24BitPixels *blockPtr, char *fileName)
 *
 * PURPOSE
 *		Attempts to load the specified photoshop 2.5 file.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *		02/20/95 (GAT) - Don't read directly into data structure as it
 *						 is padded on an SGI.
 *
 * SEE ALSO
 *
*/
PROC (void , loadPhotoshop, (BlockO24BitPixels *blockPtr, char *fileName))
{
	UBYTE		 header[4+2+6+2+4+4+2+2];
	UBYTE		**channelBufs;
	int			 r;
	WORD		 compress;
	int			 i;
	int			 line;
	long		 width;
	long		 height;
	long		 channels;
	long		 dsize;
	PSHeader	 psh;
	MEMFILE		*mf;

	mf = MEMFILE_Load (fileName);

// read the header

	if (MEMFILE_Read(mf, &header, sizeof(header)) != sizeof(header))
	{
		FailMess ("loadPhotoshop:Bad Header\n");
	}

	psh.signature = ((ULONG)header[ 0] << 24) | ((ULONG)header[ 1] << 16) | ((ULONG)header[ 2] << 8) | ((ULONG)header[ 3]);
	psh.version   = ((UWORD)header[ 4] <<  8) | ((UWORD)header[ 5]);
	psh.channels  = ((UWORD)header[12] <<  8) | ((UWORD)header[13]);
	psh.rows      = ((ULONG)header[14] << 24) | ((ULONG)header[15] << 16) | ((ULONG)header[16] << 8) | ((ULONG)header[17]);
	psh.columns   = ((ULONG)header[18] << 24) | ((ULONG)header[19] << 16) | ((ULONG)header[20] << 8) | ((ULONG)header[21]);
	psh.depth     = ((UWORD)header[22] <<  8) | ((UWORD)header[23]);
	psh.mode      = ((UWORD)header[24] <<  8) | ((UWORD)header[25]);

//	BigLong2Native(psh.signature);
//	BigWord2Native(psh.version);
//	BigWord2Native(psh.channels);
//	BigLong2Native(psh.rows);
//	BigLong2Native(psh.columns);
//	BigWord2Native(psh.depth);
//	BigWord2Native(psh.mode);

// Verify the format

	if (psh.signature != 0x38425053)
	{
		FailMess ("loadPhotoshop:Cannot identify signature (%08lx)\n", psh.signature);
	}
	if (psh.version != 1)
	{
		FailMess ("loadPhotoshop:Bad version number (%d)\n", psh.version);
	}
	if (psh.channels < 3)
	{
		FailMess ("loadPhotoshop:Wrong number of channels (%d)\n", psh.channels);
	}
	if (psh.depth != 8)
	{
		FailMess ("loadPhotoshop:Wrong depth (%d)\n", psh.depth);
	}
	if (psh.mode != 3)
	{
		FailMess ("loadPhotoshop:Wrong Mode (not RGB(%d))\n", psh.mode);
	}

	if (MEMFILE_Read(mf, &dsize, 4) != 4)
	{
		FailMess("Error reading mode data\n");
	}
	BigLong2Native (dsize);
	MEMFILE_Seek(mf, dsize, SEEK_CUR);

	if (MEMFILE_Read(mf, &dsize,4) != 4)
	{
		FailMess("Error reading resource data\n");
	}
	BigLong2Native (dsize);
	MEMFILE_Seek(mf, dsize, SEEK_CUR);

	if (MEMFILE_Read(mf, &dsize, 4) != 4)
	{
		FailMess("Error reading reserved data\n");
	}
	BigLong2Native (dsize);
	MEMFILE_Seek(mf, dsize, SEEK_CUR);

// check compression

	if (MEMFILE_Read(mf, &compress, 2) != 2)
	{
		FailMess("Error reading compression type\n");
	}
	BigWord2Native (compress);

	width    = psh.columns;
	height   = psh.rows;
	channels = psh.channels - 3;

	blockPtr->width    = width;
	blockPtr->height   = height;
	blockPtr->channels = channels;

	{
		long	channelSize;
		
		channelBufs = CHK_AllocateMemory ((channels + 3) * sizeof (UBYTE *), "channel bufs");

		channelSize = width * height;

		blockPtr->rgb = CHK_AllocateMemory (channelSize * 3, "rgb buffer");

		channelBufs[0] = blockPtr->rgb;
		channelBufs[1] = blockPtr->rgb + 1;
		channelBufs[2] = blockPtr->rgb + 2;

		if (channels)
		{
			long	i;

			blockPtr->channel = CHK_AllocateMemory (channels * sizeof (UBYTE *), "channel pointers");
			for (i = 0; i < channels; i++)
			{
				blockPtr->channel[i] = (UBYTE *)CHK_AllocateMemory (channelSize, "channel buffer");
				channelBufs[i + 3] = blockPtr->channel[i];
			}
		}
	}

// uncompress the body

	if (!compress)
	{
	// load uncompressed

		long	channel;

		for (channel = 0; channel < channels + 3; channel++)
		{
			for(line = 0; line < height; line++)
			{
				for(i = 0; i < width; i++)
				{
					if (EOF == (r = MEMFILE_getc(mf)))
					{
						FailMess("Error reading pic data\n");
					}
					*channelBufs[channel] = (UBYTE)r;
					channelBufs[channel] += (channel < 3) ? 3 : 1;
				}
			}
		}
	}
	else
	{
	// load compressed
		long	junk;
		long	channel;

		junk = psh.rows*psh.channels*2;
		MEMFILE_Seek (mf, junk, SEEK_CUR);	// Skip line size info

		for (channel = 0; channel < channels + 3; channel++)
		{
			if (unpack(channelBufs[channel], mf, width, height, (channel < 3) ? 3 : 1))
			{
				FailMess("Error reading pic data\n");
			}
		}
	}

	CHK_DeallocateMemory (channelBufs, "channel pointers");
}
PROC_END // loadPhotoshop

