/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*****************************************************************************
** EasyDBuf.h
**
** Routines for double buffering going directly to the hardware in a
** system friendly way.
**
** Routines:
**   struct Screen *CreateDBufScreen ();
**
**	returns a pointer to a 320x200x5 Intuition Screen and also
**	initializes all	that needs to be initialized to do double buffering.
**	Two global rastport pointers are setup called FrontRp and BackRp.
**	FrontRp always points to the "screen" being displayed.  BackRp
**	always points to the "screen" that is hidden or in the back.  You
**	will most likely always draw to BackRp.  FrontRp and BackRp are
**	not layered and therefore the operating system will not do any
**	clipping for you.  You may add layers and or cliprects so that the
**	OS will clip but be sure to remove them before you call
**	DeleteDBufScreen ().
**	NOTE: Screen->UserData is used by these routines.  Don't use it!
**
**   VOID DeleteDBufScreen (scr);
**	struct Screen *scr;
**
**	Closes and Deallocates all resources that were opened by
**	CreateDBufScreen ().
**
**   VOID FlipDBufScreen (scr);
**	struct Screen *scr;
**
**	Tells the Double buffering routines to flip the Front and Back
**	screens.  NOTE: The "flip" doesn't happen until the next Vertical
**      Blank or the next bottom of the screen although this routine will
**	return before that time.  It is therefore not safe to draw to
**	BackRp until after a call to WaitDBufScreen ().  It is suggested
**	that you do some calculations neccesary for your program between
**	calls to FlipDBufScreen() and WaitDBufScreen() so as to make the
**	most effective use of the time available.
**
**   VOID WaitDBufScreen (scr);
**	struct Screen *scr;
**
**	Waits for the "screens" to actually "flip".  NOTE: you must call
**	FlipDBufScreen () sometime before each call to WaitDBufScreen ()
**	or WaitDBufScreen() may never return.  You do NOT have to call
**	WaitDBufScreen() for each FlipDBufScreen () but if you don't you
**	will risk flicker when rendering to BackRp.
**
**   VOID PositionDBufScreen (scr, x, y);
**	struct Screen	*scr;
**	UWORD		x, y;
**
**	Positions the entire display in a way similar to the way Preferences
**	allows the user to adjust the position of the display.  You may call
**	this from somewhere in your program to allow the user to "set
**	preferences" or if you want to shake the screen for an effect.
**	Valid values for x and y are the values from a standard View (see
**	example TEST.c).  The default x, y on my system are x = $81, y = $2D;
**	NOTE: Any sprites on the screen will not be affected until the next
**	call to the sprite routines for that particular sprite.  (ie. The
**	Pointer/Arrow will not change position relative to the display
**	until the user moves the mouse.)
**
**   VOID SetRGB (scr, reg, color);
**	struct Screen	*scr;
**	UWORD		reg, color;
**
**      Sets color "reg" to "color".  "Reg" should be a value from 0 to 31.
**	Values out side of this will likely crash the system.  "Color" should
**	be a "standard" color value (ie. 0x0RGB) R = Red Nybble, G = Green
**	Nybble, B = Blue Nybble.
**	NOTE : Define DEBUG when debugging so that Intuition screen will
**	       have the same colors as the "copperlist".  This will of
**	       course SLOW down SetRGB by a few orders of magnitude! (PEH!)
**
**   VOID LoadRGB (scr, reg, colors, count);
**	struct Screen	*scr;
**	UWORD		reg, *colors, count;
**
**	Sets a range of colors from "reg" to ("reg" + "count" - 1).  Gets
**	the color from the UWORDs pointed to by "colors".
**
*******************************************************************************
**
**	
*******************************************************************************/

#ifndef EASYDBUF_H
#define EASYDBUF_H TRUE

#ifndef INTUITION_INTUITION_H
#include <intuition/intuition.h>
#endif

typedef struct dbufscreentrk {
	struct	BitMap	*FrontBitMapPtr;
	struct	BitMap	*BackBitMapPtr;
	struct	BitMap	BackBitMap;
} DBufScreenTrkType;

struct	Screen	*CreateDBufScreen (void);
VOID		DeleteDBufScreen (struct Screen *src);
VOID		RectifyDBufScreen (void);
VOID		FlipDBufScreen (void);
VOID		WaitDBufScreen (void);
VOID		PositionDBufScreen (UWORD x, UWORD y);
VOID		SetRGB (UWORD reg, UWORD color);
VOID		LoadRGB (UWORD reg, UWORD *colors, UWORD count);
VOID		SetRGBBoth (UWORD reg, UWORD color);
VOID		LoadRGBBoth (UWORD reg, UWORD *colors, UWORD count);
VOID		RestoreView (void);

extern	struct	RastPort	*FrontRp;
extern	struct	RastPort	*BackRp;
extern  ULONG				 VBTimer;

#endif

