/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * ALLOC.C
 *
 * PROGRAMMER : Dan Chang
 *    VERSION : 00.000
 *    CREATED : 07/21/94
 *   MODIFIED : 03/28/95
 *       TABS : 04 07
 *
 * DESCRIPTION
 *		Debugging heap routines. Based upon No Bugs!: delivering error-free
 *		code in C and C++ by David Thielen, Addison-Wesley, 1992, pages 69..82.
 *
 * HISTORY
 *		03/31/92 (D T) - Created.
 *		08/05/94 (dcc) - Remove typos, make compile with BCC 3.1.
 *		10/03/94 (dcc) - Enhance error reporting, support MEM_malloc(0),
 *							  add checksum.
 *		10/05/94 (dcc) - Added MEM_farmalloc(), MEM_farcalloc(), MEM_farfree()
 *							  and MEM_farrealloc().
 *		10/08/94 (dcc) - Added MEM_SetShowCalls().
 *		10/10/94 (dcc) - Added MEM_SetCheckHeapRate(), MEM_SetShowOOM().
 *		03/28/95 (dcc) - Since now part of libraries, remove #if DEBUG.
 *
*/

#include <echidna/platform.h>
#include "switches.h"						// this must precede <echidna/alloc.h>

#include <echidna/alloc.h>

#if (__MSDOS16__ && __BORLANDC__)
#if (!defined(__COMPACT__) && !defined(__LARGE__) && !defined(__HUGE__))
#error No support for this memory model.
#endif
#endif

#include <alloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <mem.h>
#include <values.h>

/**************************** C O N S T A N T S ***************************/

#if (sizeof(unsigned) == 2)
#define	ALLOC_ID				0x1234U		// unsigned preceding tracking pointers
#define	ALLOC_CHK1			0x5A96U		// unsigned preceding malloc'ed memory
#define	ALLOC_CHK2			0xA569U		// unsigned following malloc'ed memory
#define	ALLOC_CHK3			0xEF			// byte to fill freed mem

#if __MSDOS16__
#define	uAllocMax			0xFFFFU
#else		/* !__MSDOS16__ */
#define	uAllocMax			(0xFFFEU-sizeof(ALLOC)-sizeof(unsigned))
#endif	/* !__MSDOS16__ */
#define	ulAllocMax			(0xFFFFFFFEUL-sizeof(ALLOC)-sizeof(ULONG))
#elif (sizeof(unsigned) == 4)
#define	ALLOC_ID				0x1234DC78U	// unsigned preceding tracking pointers
#define	ALLOC_CHK1			0x5A9665A9U	// unsigned preceding malloc'ed memory
#define	ALLOC_CHK2			0xA569569AU	// unsigned following malloc'ed memory
#define	ALLOC_CHK3			0xEF			// byte to fill freed mem

#define	uAllocMax			(0xFFFFFFFEU-sizeof(ALLOC)-sizeof(unsigned))
#endif


/******************************** T Y P E S *******************************/

typedef struct
{
	struct S_ALLOC far32	*pPrevAlloc;	// points to prev alloc
	struct S_ALLOC far32	*pNextAlloc;	// points to next alloc
	char						*pFileName;		// points to file
	int						 iFileLine;		// points to line number
	ULONG						 ulSize;			// size of allocated block
} AllocInfo;

typedef struct S_ALLOC
{
	unsigned					 uCheckSum;		// checksum of following AllocInfo
	AllocInfo				 a;				// prev, next, file, line #
	unsigned					 uChk1;			// ALLOC_CHK1
} ALLOC;

/****************************** G L O B A L S *****************************/

static BOOL		fNoisy = FALSE;
static int		iCheckHeapRate = 1;
static BOOL		fShowOOM = TRUE;
static int		cCheckHeap = 0;

static ALLOC	__AllocHeap = {0, {NULL, NULL, NULL, 0, 0}, 0};

/******************************* M A C R O S ******************************/

#define DebugPrintf printf

/***************************** R O U T I N E S ****************************/


#if (sizeof(unsigned) == 2)
/*********************************************************************
 *
 * MEM_ulmemcpy
 *
 * PURPOSE
 *		copy <ul> bytes from <psrc> to <pdst>.
 *
 * INPUT
 *		void far32 *pdst			: pointer to destination to copy bytes to
 *		const void far32 *psrc	: pointer to source to copy bytes from
 *		ULONG ul						: number of bytes to copy
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		Returns *pdst.
 *
 * HISTORY
 *		08/05/94 Friday (dcc) - created.
 *
*/
static void far32 *MEM_ulmemcpy(void far32 *pdst, const void far32 *psrc, ULONG ul)
{
	char huge32 *pubdst = pdst;
	const char huge32 *pubsrc = psrc;

	while (ul > 0)
	{
		size_t i;

		if (ul > MAXINT)
			i = (size_t) MAXINT;
		else
			i = (size_t) ul;

#if __BORLANDC__
		_fmemcpy(pubdst, pubsrc, i);
#else
		memcpy(pubdst, pubsrc, i);
#endif
		ul -= i;
		pubdst += i;
		pubsrc += i;
	}
	return pdst;

} /* MEM_ulmemcpy */


/*********************************************************************
 *
 * MEM_ulmemset
 *
 * PURPOSE
 *		Set <ul> bytes pointed to by <s> to <c>.
 *
 * INPUT
 *		void far32 *s		: pointer to memory to set
 *		int c					: value to set memory to
 *		ULONG ul				: number of bytes to set
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		Returns <s>.
 *
 * HISTORY
 *		08/05/94 Friday (dcc) - created.
 *
*/
static void far32 *MEM_ulmemset(void far32 *s, int c, ULONG ul)
{
	char huge32 *pub = s;

	while (ul > 0)
	{
		size_t i;

		if (ul > MAXINT)
			i = (size_t) MAXINT;
		else
			i = (size_t) ul;

#if __BORLANDC__
		_fmemset(pub, c, i);
#else
		memset(pub, c, i);
#endif
		ul -= i;
		pub += i;
	}
	return s;
} /* MEM_ulmemset */
#elif (sizeof(unsigned) == 4)
#define MEM_ulmemcpy						memcpy
#define MEM_ulmemset						memset
#endif


/*********************************************************************
 *
 * ComputeCheckSum
 *
 * PURPOSE
 *		Compute the checksum for the values in pAlloc->a, and
 *		store the results in pAlloc->uCheckSum.
 *
 * INPUT
 *		ALLOC huge32 *pAlloc	: pointer to ALLOC to compute & set checksum
 *
 * ASSUMES
 *		pAlloc->a all contain good values.
 *
 * RETURN VALUE
 *		NONE.
 *
 * HISTORY
 *		10/03/94 (dcc) - created.
 *
*/
static void ComputeCheckSum(ALLOC huge32 *pAlloc)
{
	/* NULLTRAPOFF */
	int i;
	UBYTE far32 *pub = (UBYTE far32 *) &pAlloc->a;
	unsigned uCheckSum = 0U;

	for (i = 0; i < sizeof(AllocInfo); i++)
	{
		uCheckSum += *pub;
		pub++;
	}
	pAlloc->uCheckSum = uCheckSum;
	/* NULLTRAPON */
} /* ComputeCheckSum */


/*********************************************************************
 *
 * FCheckSumOK
 *
 * PURPOSE
 *		Verify the checksum stored in pAlloc->uCheckSum
 *		for the values in pAlloc->a.
 *
 * INPUT
 *		ALLOC huge32 *pAlloc	: pointer to ALLOC to verify
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		TRUE if checksum is OK.
 *
 * HISTORY
 *		10/03/94 (dcc) - created.
 *
*/
static BOOL FCheckSumOK(ALLOC huge32 *pAlloc)
{
	/* NULLTRAPOFF */
	int i;
	UBYTE far32 *pub = (UBYTE far32 *) &pAlloc->a;
	unsigned uCheckSum = 0U;

	for (i = 0; i < sizeof(AllocInfo); i++)
	{
		uCheckSum += *pub;
		pub++;
	}
	return (pAlloc->uCheckSum == uCheckSum);
	/* NULLTRAPON */
} /* FCheckSumOK */


/*********************************************************************
 *
 * OpenAllocPrintf
 *
 * PURPOSE
 *		Set data items before first AllocPrintf().
 *
 * INPUT
 *		char *sz				: name of calling routine
 *		char *pFile			: file name of module calling <sz>()
 *		int iLine			: line number in module calling <sz>()
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		
 *
 * HISTORY
 *		10/03/94 (dcc) - created.
 *
*/
static char	*szFuncPrint;
static char	*pFilePrint;
static int	 iLinePrint;
static BOOL	 fFirstPrint;
static int	 iPrintLevel = 0;

static void OpenAllocPrintf(char *sz, char *pFile, int iLine)
{
	if (iPrintLevel == 0)
	{
		szFuncPrint	= sz;
		pFilePrint	= pFile;
		iLinePrint	= iLine;
		fFirstPrint	= TRUE;
		if (fNoisy)
			DebugPrintf("%s %d: %s", pFilePrint, iLinePrint, szFuncPrint);
	}
	iPrintLevel++;
} /* OpenAllocPrintf */


/*********************************************************************
 *
 * _AllocPrintf
 *
 * PURPOSE
 *		Print the message.
 *
 * INPUT
 *		ALLOC far32 *pAlloc	: pointer to ALLOC to print
 *		char *pStr				: message to print
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		NONE.
 *
 * HISTORY
 *		03/31/92 (D T) - created.
 *		10/03/94 (dcc) - enhanced error reporting.
 *
*/
static void _AllocPrintf(ALLOC far32 *pAlloc, char *pStr)
{
	if (fFirstPrint)
	{
		if (!fNoisy)
			DebugPrintf("%s %d: %s", pFilePrint, iLinePrint, szFuncPrint);

		DebugPrintf("\n{\n");
		fFirstPrint = FALSE;
	}

	if (pAlloc != NULL)
	{
		if (FCheckSumOK(pAlloc))
			DebugPrintf(" %s %s %d: %p: %lu bytes\n", pStr,
							pAlloc->a.pFileName, pAlloc->a.iFileLine, pAlloc+1,
							pAlloc->a.ulSize);
		else
			DebugPrintf(" %s corrupted block: %p\n", pStr, pAlloc+1);
	}
	else
	{
/*GOOD*/		DebugPrintf(" %s\n", pStr);
	}
} /* _AllocPrintf */


/*********************************************************************
 *
 * CloseAllocPrintf
 *
 * PURPOSE
 *		If calls were made to _AllocPrintf(), then print a closing '}'.
 *
 * INPUT
 *		BOOL fRtnValue		: if TRUE, print a return value, else ignore <pv>
 *		void far32 *pv		: the return value
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		NONE.
 *
 * HISTORY
 *		10/03/94 (dcc) - created.
 *
*/
static void CloseAllocPrintf(BOOL fRtnValue, void far32 *pv)
{
	iPrintLevel--;
	if (iPrintLevel == 0)
	{
		if (!fFirstPrint)
		{
			DebugPrintf("}");
		}
		if (fNoisy && fRtnValue)
		{
			DebugPrintf(" returns %p\n", pv);
			return;
		}
		if (!fFirstPrint)
		{
			DebugPrintf("\n");
		}
	}
} /* CloseAllocPrintf */


/*********************************************************************
 *
 * __FHeapOK
 *
 * PURPOSE
 *		Verify that the heap is not corrupt.
 *
 * INPUT
 *		BOOL fShowOK		: if TRUE, show every allocated block
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		TRUE if heap is OK, FALSE if heap is corrupted.
 *
 * HISTORY
 *		03/31/92 (D T) - created.
 *		10/03/94 (dcc) - enhanced error reporting.
 *
*/
static BOOL __FHeapOK(BOOL fShowOK)
{
	char *pStr;
	ALLOC far32 *pAlloc;

	/* Walk the heap - check CheckSum, begin and end check words. */

	if ((pAlloc = __AllocHeap.a.pNextAlloc) == NULL)
	{
		if (fShowOK)
			_AllocPrintf(NULL, "Empty heap.");
		return TRUE;
	}

	while (pAlloc != &__AllocHeap)
	{
		if (!FCheckSumOK(pAlloc))
		{
			_AllocPrintf(pAlloc, "Heap Corrupted;");
			return FALSE;
		}
		pStr = NULL;

		if (pAlloc->uChk1 != ALLOC_CHK1)
			pStr = "Underwrite error";

		if (*((unsigned huge32 *) (((char huge32 *) pAlloc) + sizeof(ALLOC) +
				pAlloc->a.ulSize)) != ALLOC_CHK2)
		{
			if (pStr)
				pStr = "Under & Overwrite error";
			else
				pStr = "OverWrite error";
		}

		if (pStr)
			_AllocPrintf(pAlloc, pStr);
		else
			if (fShowOK)
				_AllocPrintf(pAlloc, "Good block");

		pAlloc = pAlloc->a.pNextAlloc;
	}
	return TRUE;
} /* __FHeapOK */


/*********************************************************************
 *
 * FMaybeCheckHeap
 *
 * PURPOSE
 *		Increment <cCheckHeap>. If <cCheckHeap> >= <iCheckHeapRate>,
 *		then call _FHeapOK(FALSE) and return the result of that call.
 *		Otherwise just return TRUE.
 *
 * INPUT
 *		NONE.
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		FALSE if heap is corrupted.
 *
 * HISTORY
 *		10/10/94 (dcc) - created.
 *
*/
static BOOL FMaybeCheckHeap(void)
{
	cCheckHeap++;
	if (cCheckHeap >= iCheckHeapRate)
	{
		cCheckHeap = 0;
		return __FHeapOK(FALSE);
	}
	return TRUE;
} /* FMaybeCheckHeap */


/*********************************************************************
 *
 * __SetShowCalls
 *
 * PURPOSE
 *		Set <fNoisy>.
 *
 * INPUT
 *		BOOL f				: if TRUE, prints a message with every
 *								  MEM_calloc(), MEM_free(), etc.
 *		char *pFile			: file name of module calling __SetShowCalls()
 *		int iLine			: line number in module calling __SetShowCalls()
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		NONE.
 *
 * HISTORY
 *		10/08/94 (dcc) - created.
 *
*/
void __SetShowCalls(BOOL f, char *pFile, int iLine)
{
	if (fNoisy || f)
		DebugPrintf("%s %d: MEM_SetShowCalls(%d)\n", pFile, iLine, f);

	fNoisy = f;
} /* __SetShowCalls */


/*********************************************************************
 *
 * __SetCheckHeapRate
 *
 * PURPOSE
 *		Set how frequently the heap is checked. If passed in 1, then
 *		the heap is verified with each MEM_calloc(), MEM_free(), etc.
 *
 * INPUT
 *		int i					: how frequently should be heap be checked
 *		char *pFile			: file name of module calling __SetCheckHeapRate()
 *		int iLine			: line number in module calling __SetCheckHeapRate()
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		
 *
 * HISTORY
 *		10/10/94 (dcc) - created.
 *
*/
void __SetCheckHeapRate(int i, char *pFile, int iLine)
{
	if (fNoisy)
		DebugPrintf("%s %d: MEM_SetCheckHeapRate(%d)\n", pFile, iLine, i);

	iCheckHeapRate = i;
} /* __SetCheckHeapRate */


/*********************************************************************
 *
 * __GetShowOOM
 *
 * PURPOSE
 *		Return current <fShowOOM>.
 *
 * INPUT
 *		char *pFile			: file name of module calling __GetShowOOM()
 *		int iLine			: line number in module calling __GetShowOOM()
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		Returns <fShowOOM>.
 *
 * HISTORY
 *		10/10/94 (dcc) - created.
 *
*/
BOOL __GetShowOOM(char *pFile, int iLine)
{
	if (fNoisy)
		DebugPrintf("%s %d: MEM_GetShowOOM() returns %d\n", pFile, iLine, fShowOOM);

	return fShowOOM;
} /* __GetShowOOM */


/*********************************************************************
 *
 * __SetShowOOM
 *
 * PURPOSE
 *		Set if routines should print an error on OOM condition.
 *
 * INPUT
 *		BOOL f				: if TRUE, prints a message when OOM.
 *		char *pFile			: file name of module calling __SetShowOOM()
 *		int iLine			: line number in module calling __SetShowOOM()
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		NONE.
 *
 * HISTORY
 *		10/10/94 (dcc) - created.
 *
*/
void __SetShowOOM(BOOL f, char *pFile, int iLine)
{
	if (fNoisy)
		DebugPrintf("%s %d: MEM_SetShowOOM(%d)\n", pFile, iLine, f);

	fShowOOM = f;
} /* __SetShowOOM */


/*********************************************************************
 *
 * __ShowHeap
 *
 * PURPOSE
 *		Print out the heap.
 *
 * INPUT
 *		char *pFile			: file name of module calling __ShowHeap()
 *		int iLine			: line number in module calling __ShowHeap()
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		NONE.
 *
 * HISTORY
 *		10/03/94 (dcc) - created.
 *
*/
void __ShowHeap(char *pFile, int iLine)
{
	OpenAllocPrintf("MEM_ShowHeap()", pFile, iLine);

	__FHeapOK(TRUE);

	CloseAllocPrintf(FALSE, NULL);
} /* __ShowHeap */


/*********************************************************************
 *
 * __malloc
 *
 * PURPOSE
 *		Allocate memory to store <uSize> bytes.
 *
 * INPUT
 *		size_t uSize		: number of bytes to allocate
 *		char *pFile			: file name of module calling __malloc()
 *		int iLine			: line number in module calling __malloc()
 *		BOOL fCheckHeap	: if TRUE, verify that the heap is still OK
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		Pointer to allocated memory, or NULL if OOM.
 *
 * HISTORY
 *		03/31/92 (D T) - created.
 *		10/03/94 (dcc) - return NULL if uSize == 0, enhanced error reporting.
 *
*/
void *__malloc(size_t uSize, char *pFile, int iLine, BOOL fCheckHeap)
{
	ALLOC far32 *pAlloc;
	ALLOC sAlloc;
	char sz[256];

	sprintf(sz, "MEM_malloc(%u)", uSize);
	OpenAllocPrintf(sz, pFile, iLine);

#if (sizeof(size_t) == 2)
#if !__MSDOS16__
	if (uSize > uAllocMax)
	{
		char sz[256];

		sprintf(sz, "Error: unable to allocate more than %u bytes.", uAllocMax);
		_AllocPrintf(NULL, sz);
		if (fCheckHeap)
			FMaybeCheckHeap();
		CloseAllocPrintf(TRUE, NULL);
		return NULL;
	}
#endif	/* !__MSDOS16__ */
#elif (sizeof(size_t) == 4)
	if (uSize > uAllocMax)
	{
		char sz[256];

		sprintf(sz, "Error: unable to allocate more than %u bytes.", uAllocMax);
		_AllocPrintf(NULL, sz);
		if (fCheckHeap)
			FMaybeCheckHeap();
		CloseAllocPrintf(TRUE, NULL);
		return NULL;
	}
#endif

	/* See if the heap is still OK. */

	if (fCheckHeap)
	{
		if (!FMaybeCheckHeap())
		{
			CloseAllocPrintf(TRUE, NULL);
			return NULL;
		}
	}

	if (uSize == 0)
	{
		CloseAllocPrintf(TRUE, NULL);
		return NULL;
	}

#if __MSDOS16__
	if ((pAlloc = farmalloc((ULONG)uSize+sizeof(ALLOC)+sizeof(unsigned))) == NULL)
#else		/* !__MSDOS16__ */
	if ((pAlloc = malloc(uSize+sizeof(ALLOC)+sizeof(unsigned))) == NULL)
#endif	/* !__MSDOS16__ */
	{
		if (fShowOOM)
		{
			_AllocPrintf(NULL, "Error: out of memory.");
			__FHeapOK(TRUE);
		}
		CloseAllocPrintf(TRUE, NULL);
		return NULL;
	}

	sAlloc.a.pFileName	= pFile;
	sAlloc.a.iFileLine	= iLine;
	sAlloc.a.ulSize		= uSize;
	sAlloc.uChk1			= ALLOC_CHK1;

	/* Insert it into the linked list. */

	if (__AllocHeap.a.pPrevAlloc == NULL)
	{
		sAlloc.a.pPrevAlloc = &__AllocHeap;
		__AllocHeap.a.pNextAlloc = pAlloc;
	}
	else
	{
		sAlloc.a.pPrevAlloc = __AllocHeap.a.pPrevAlloc;
		(__AllocHeap.a.pPrevAlloc)->a.pNextAlloc = pAlloc;
		ComputeCheckSum(__AllocHeap.a.pPrevAlloc);
	}
	sAlloc.a.pNextAlloc = &__AllocHeap;
	__AllocHeap.a.pPrevAlloc = pAlloc;

	ComputeCheckSum(&sAlloc);
	ComputeCheckSum(&__AllocHeap);

	// put the struct contents in
	*pAlloc++ = sAlloc;

	/* Zero it out and put the check word at the end. */

	MEM_ulmemset(pAlloc, 0, uSize);
	*((unsigned huge32 *) ((char huge32 *) pAlloc + uSize)) = ALLOC_CHK2;

	CloseAllocPrintf(TRUE, pAlloc);
	return (void *) pAlloc;
} /* __malloc */


/*********************************************************************
 *
 * __calloc
 *
 * PURPOSE
 *		Allocate memory to store <cItems> of size <uSize>.
 *
 * INPUT
 *		size_t cItems		: number of items to allocate
 *		size_t uSize		: number of bytes per item
 *		char *pFile			: file name of module calling __calloc()
 *		int iLine			: line number in module calling __calloc()
 *		BOOL fCheckHeap	: if TRUE, verify that the heap is still OK
 *
 * ASSUMES
 *		That __malloc() will zero the allocated buffer.
 *
 * RETURN VALUE
 *		Pointer to allocated memory, or NULL if OOM.
 *
 * HISTORY
 *		03/31/92 (D T) - created.
 *		10/03/94 (dcc) - return NULL if uSize == 0, enhanced error reporting.
 *
*/
void *__calloc(size_t cItems, size_t uSize, char *pFile, int iLine, BOOL fCheckHeap)
{
	void *pv;
	char sz[256];

	sprintf(sz, "MEM_calloc(%u, %u)", cItems, uSize);
	OpenAllocPrintf(sz, pFile, iLine);

#if (sizeof(size_t) == 2)
	if ((ULONG) cItems * uSize > uAllocMax)
	{
		char sz[256];

		sprintf(sz, "Error: unable to allocate more than %u bytes.", uAllocMax);
		_AllocPrintf(NULL, sz);
		if (fCheckHeap)
			FMaybeCheckHeap();
		CloseAllocPrintf(TRUE, NULL);
		return NULL;
	}
#elif (sizeof(size_t) == 4)
	if ((ULONG) cItems * uSize > uAllocMax)
	{
		char sz[256];

		sprintf(sz, "Error: unable to allocate more than %u bytes.", uAllocMax);
		_AllocPrintf(NULL, sz);
		if (fCheckHeap)
			FMaybeCheckHeap();
		CloseAllocPrintf(TRUE, NULL);
		return NULL;
	}
#endif

	pv = __malloc(cItems * uSize, pFile, iLine, fCheckHeap);
	CloseAllocPrintf(TRUE, pv);
	return pv;
} /* __calloc */


/*********************************************************************
 *
 * __free
 *
 * PURPOSE
 *		Free the memory <pMem> allocated by __calloc(), __malloc() or
 *		__realloc().
 *
 * INPUT
 *		void *pMem			: memory block to free
 *		char *pFile			: file name of module calling __free()
 *		int iLine			: line number in module calling __free()
 *		BOOL fCheckHeap	: if TRUE, verify that the heap is still OK
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		NONE.
 *
 * HISTORY
 *		03/31/92 (D T) - created.
 *
*/
void __free(void *pMem, char *pFile, int iLine, BOOL fCheckHeap)
{
	ALLOC far32 *pAlloc;
	char sz[256];

	sprintf(sz, "MEM_free(%p)", pMem);
	OpenAllocPrintf(sz, pFile, iLine);

	/* Doing a free on NULL pointer is NOT an error. */

	if (pMem == NULL)
	{
		CloseAllocPrintf(FALSE, NULL);
		return;
	}

	pAlloc = ((ALLOC far32 *) pMem) - 1;		/* get the struct */

	/* See if the heap is still okay. */

	/* Is structure still valid? */

	if (!FCheckSumOK(pAlloc))
	{
		_AllocPrintf(pAlloc, "Passed in");
		if (fCheckHeap)
			FMaybeCheckHeap();
		CloseAllocPrintf(FALSE, NULL);
		return;
	}

	if (fCheckHeap)
	{
		if (!FMaybeCheckHeap())
		{
			CloseAllocPrintf(FALSE, NULL);
			return;
		}
	}

	/* Take it out of the list. */

	if (pAlloc->a.pPrevAlloc == pAlloc->a.pNextAlloc)
	{
		__AllocHeap.a.pPrevAlloc = __AllocHeap.a.pNextAlloc = NULL;
		ComputeCheckSum(&__AllocHeap);
	}
	else
	{
		(pAlloc->a.pPrevAlloc)->a.pNextAlloc = pAlloc->a.pNextAlloc;
		ComputeCheckSum(pAlloc->a.pPrevAlloc);
		(pAlloc->a.pNextAlloc)->a.pPrevAlloc = pAlloc->a.pPrevAlloc;
		ComputeCheckSum(pAlloc->a.pNextAlloc);
	}

	/* Fill up the memory in case it's accessed. */

	MEM_ulmemset(pAlloc, ALLOC_CHK3, pAlloc->a.ulSize + sizeof(ALLOC)+sizeof(unsigned));
#if __MSDOS16__
	farfree((void *) pAlloc);
#else		/* !__MSDOS16__ */
	free((void *) pAlloc);
#endif	/* !__MSDOS16__ */

	CloseAllocPrintf(FALSE, NULL);
} /* __free */


/*********************************************************************
 *
 * __realloc
 *
 * PURPOSE
 *		Free the memory pointed to by <pMem>, allocate a new block
 *		of memory of <uSize> bytes and return a pointer to the new
 *		block of memory.
 *
 *		This routines forces the returned pointer to be different
 *		then <pMem> to assist in debugging references to pointers
 *		that are no longer valid.
 *
 * INPUT
 *		void *pMem			: memory block to reallocate
 *		size_t uSize		: number of bytes to reallocate
 *		char *pFile			: file name of module calling __reallocate()
 *		int iLine			: line number in module calling __reallocate()
 *		BOOL fCheckHeap	: if TRUE, verify that the heap is still OK
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		Pointer to reallocated memory, or NULL if OOM.
 *
 * HISTORY
 *		03/31/92 (D T) - created.
 *		10/03/94 (dcc) - return NULL if uSize == 0, enhanced error reporting.
 *
*/
void *__realloc(void *pMem, size_t uSize, char *pFile, int iLine, BOOL fCheckHeap)
{
	ALLOC *pAllocOld, *pAllocNew;
	char sz[256];

	if (!pMem)
	{
		void *pv;

		sprintf(sz, "MEM_realloc(%p, %u)", pMem, uSize);
		OpenAllocPrintf(sz, pFile, iLine);

		if (uSize == 0)
			pv = NULL;
		else
			pv = __malloc(uSize, pFile, iLine, fCheckHeap);
		CloseAllocPrintf(TRUE, pv);
		return pv;
	}

	pAllocOld = (ALLOC *) (((ALLOC huge32 *) pMem) - 1);	/* get the struct */

	sprintf(sz, "MEM_realloc(/* %s %d: */ %p, %u)",
				pAllocOld->a.pFileName, pAllocOld->a.iFileLine, pMem, uSize);
	OpenAllocPrintf(sz, pFile, iLine);

	/* Is structure still valid? */

	if (!FCheckSumOK(pAllocOld))
	{
		_AllocPrintf(pAllocOld, "Passed in");
		if (fCheckHeap)
			FMaybeCheckHeap();
		CloseAllocPrintf(TRUE, NULL);
		return NULL;
	}

//	/* Fill out values for __CheckHeap() and _AllocPrintf(). */

//	pAllocOld->a.pFileName = pFile;
//	pAllocOld->a.iFileLine = iLine;

	/* See if the heap is still okay. */

	if (fCheckHeap)
	{
		if (!FMaybeCheckHeap())
		{
			CloseAllocPrintf(TRUE, NULL);
			return NULL;
		}
	}

	/* Get the new memory block of <uSize> bytes. */

	if (uSize == 0)
	{
		pAllocNew = NULL;
	}
	else
	{
		if ((pAllocNew = __malloc(uSize, pFile, iLine, FALSE)) != NULL)
		{
			/* copy contents across (no need to zero extra, __malloc() did it). */

			MEM_ulmemcpy(pAllocNew, pMem, min(uSize, pAllocOld->a.ulSize));
		}
	}
	__free(pMem, pFile, iLine, FALSE);	// don't check heap again!

	CloseAllocPrintf(TRUE, pAllocNew);
	return (void *) pAllocNew;
} /* __realloc */


#if __MSDOS16__
/*********************************************************************
 *
 * __farmalloc
 *
 * PURPOSE
 *		Allocate memory to store <ulSize> bytes.
 *
 * INPUT
 *		ULONG ulSize		: number of bytes to allocate
 *		char *pFile			: file name of module calling __farmalloc()
 *		int iLine			: line number in module calling __farmalloc()
 *		BOOL fCheckHeap	: if TRUE, verify that the heap is still OK
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		Pointer to allocated memory, or NULL if OOM.
 *
 * HISTORY
 *		03/31/92 (D T) - created.
 *		10/03/94 (dcc) - return NULL if uSize == 0, enhanced error reporting.
 *
*/
void far32 *__farmalloc(ULONG ulSize, char *pFile, int iLine, BOOL fCheckHeap)
{
	ALLOC far32 *pAlloc;
	ALLOC sAlloc;
	char sz[256];

	sprintf(sz, "MEM_farmalloc(%lu)", ulSize);
	OpenAllocPrintf(sz, pFile, iLine);

	if (ulSize > ulAllocMax)
	{
		char sz[256];

		sprintf(sz, "Error: unable to allocate more than %lu bytes.", ulAllocMax);
		_AllocPrintf(NULL, sz);
		if (fCheckHeap)
			FMaybeCheckHeap();
		CloseAllocPrintf(TRUE, NULL);
		return NULL;
	}

	/* See if the heap is still OK. */

	if (fCheckHeap)
	{
		if (!FMaybeCheckHeap())
		{
			CloseAllocPrintf(TRUE, NULL);
			return NULL;
		}
	}

	if (ulSize == 0)
	{
		CloseAllocPrintf(TRUE, NULL);
		return NULL;
	}

	if ((pAlloc = farmalloc(ulSize+sizeof(ALLOC)+sizeof(unsigned))) == NULL)
	{
		if (fShowOOM)
		{
			_AllocPrintf(NULL, "Error: out of memory.");
			__FHeapOK(TRUE);
		}
		CloseAllocPrintf(TRUE, NULL);
		return NULL;
	}

	sAlloc.a.pFileName	= pFile;
	sAlloc.a.iFileLine	= iLine;
	sAlloc.a.ulSize		= ulSize;
	sAlloc.uChk1			= ALLOC_CHK1;

	/* Insert it into the linked list. */

	if (__AllocHeap.a.pPrevAlloc == NULL)
	{
		sAlloc.a.pPrevAlloc = &__AllocHeap;
		__AllocHeap.a.pNextAlloc = pAlloc;
	}
	else
	{
		sAlloc.a.pPrevAlloc = __AllocHeap.a.pPrevAlloc;
		(__AllocHeap.a.pPrevAlloc)->a.pNextAlloc = pAlloc;
		ComputeCheckSum(__AllocHeap.a.pPrevAlloc);
	}
	sAlloc.a.pNextAlloc = &__AllocHeap;
	__AllocHeap.a.pPrevAlloc = pAlloc;

	ComputeCheckSum(&sAlloc);
	ComputeCheckSum(&__AllocHeap);

	// put the struct contents in
	*pAlloc++ = sAlloc;

	/* Zero it out and put the check word at the end. */

	MEM_ulmemset(pAlloc, 0, ulSize);
	*((unsigned huge32 *) ((char huge32 *) pAlloc + ulSize)) = ALLOC_CHK2;

	CloseAllocPrintf(TRUE, pAlloc);
	return (void far32 *) pAlloc;
} /* __farmalloc */


/*********************************************************************
 *
 * __farcalloc
 *
 * PURPOSE
 *		Allocate memory to store <cItems> of size <uSize>.
 *
 * INPUT
 *		ULONG cItems		: number of items to allocate
 *		ULONG ulSize		: number of bytes per item
 *		char *pFile			: file name of module calling __farcalloc()
 *		int iLine			: line number in module calling __farcalloc()
 *		BOOL fCheckHeap	: if TRUE, verify that the heap is still OK
 *
 * ASSUMES
 *		That __farmalloc() will zero the allocated buffer.
 *
 * RETURN VALUE
 *		Pointer to allocated memory, or NULL if OOM.
 *
 * HISTORY
 *		03/31/92 (D T) - created.
 *		10/03/94 (dcc) - return NULL if uSize == 0, enhanced error reporting.
 *
*/
void far32 *__farcalloc(ULONG cItems, ULONG ulSize, char *pFile, int iLine, BOOL fCheckHeap)
{
	void far32 *pv;
	char sz[256];

	sprintf(sz, "MEM_farcalloc(%lu, %lu)", cItems, ulSize);
	OpenAllocPrintf(sz, pFile, iLine);

	if ( cItems * ulSize > ulAllocMax)
	{
		char sz[256];

		sprintf(sz, "Error: unable to allocate more than %lu bytes.", ulAllocMax);
		_AllocPrintf(NULL, sz);
		if (fCheckHeap)
			FMaybeCheckHeap();
		CloseAllocPrintf(TRUE, NULL);
		return NULL;
	}

	pv = __farmalloc(cItems * ulSize, pFile, iLine, fCheckHeap);
	CloseAllocPrintf(TRUE, pv);
	return pv;
} /* __farcalloc */


/*********************************************************************
 *
 * __farfree
 *
 * PURPOSE
 *		Free the memory <pMem> allocated by __farcalloc(), __farmalloc()
 *		or __farrealloc().
 *
 * INPUT
 *		void far32 *pMem	: memory block to free
 *		char *pFile			: file name of module calling __farfree()
 *		int iLine			: line number in module calling __farfree()
 *		BOOL fCheckHeap	: if TRUE, verify that the heap is still OK
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		NONE.
 *
 * HISTORY
 *		03/31/92 (D T) - created.
 *
*/
void __farfree(void far32 *pMem, char *pFile, int iLine, BOOL fCheckHeap)
{
	ALLOC far32 *pAlloc;
	char sz[256];

	sprintf(sz, "MEM_farfree(%p)", pMem);
	OpenAllocPrintf(sz, pFile, iLine);

	/* Doing a free on NULL pointer is NOT an error. */

	if (pMem == NULL)
	{
		CloseAllocPrintf(FALSE, NULL);
		return;
	}

	pAlloc = ((ALLOC far32 *) pMem) - 1;		/* get the struct */

	/* See if the heap is still okay. */

	/* Is structure still valid? */

	if (!FCheckSumOK(pAlloc))
	{
		_AllocPrintf(pAlloc, "Passed in");
		if (fCheckHeap)
			FMaybeCheckHeap();
		CloseAllocPrintf(FALSE, NULL);
		return;
	}

	if (fCheckHeap)
	{
		if (!FMaybeCheckHeap())
		{
			CloseAllocPrintf(FALSE, NULL);
			return;
		}
	}

	/* Take it out of the list. */

	if (pAlloc->a.pPrevAlloc == pAlloc->a.pNextAlloc)
	{
		__AllocHeap.a.pPrevAlloc = __AllocHeap.a.pNextAlloc = NULL;
		ComputeCheckSum(&__AllocHeap);
	}
	else
	{
		(pAlloc->a.pPrevAlloc)->a.pNextAlloc = pAlloc->a.pNextAlloc;
		ComputeCheckSum(pAlloc->a.pPrevAlloc);
		(pAlloc->a.pNextAlloc)->a.pPrevAlloc = pAlloc->a.pPrevAlloc;
		ComputeCheckSum(pAlloc->a.pNextAlloc);
	}

	/* Fill up the memory in case it's accessed. */

	MEM_ulmemset(pAlloc, ALLOC_CHK3, pAlloc->a.ulSize + sizeof(ALLOC)+sizeof(unsigned));
	farfree((void far32 *) pAlloc);

	CloseAllocPrintf(FALSE, NULL);
} /* __farfree */


/*********************************************************************
 *
 * __farrealloc
 *
 * PURPOSE
 *		Free the memory pointed to by <pMem>, allocate a new block
 *		of memory of <uSize> bytes and return a pointer to the new
 *		block of memory.
 *
 *		This routines forces the returned pointer to be different
 *		then <pMem> to assist in debugging references to pointers
 *		that are no longer valid.
 *
 * INPUT
 *		void far32 *pMem	: memory block to reallocate
 *		ULONG ulSize		: number of bytes to reallocate
 *		char *pFile			: file name of module calling __farreallocate()
 *		int iLine			: line number in module calling __farreallocate()
 *		BOOL fCheckHeap	: if TRUE, verify that the heap is still OK
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		Pointer to reallocated memory, or NULL if OOM.
 *
 * HISTORY
 *		03/31/92 (D T) - created.
 *		10/03/94 (dcc) - return NULL if uSize == 0, enhanced error reporting.
 *
*/
void far32 *__farrealloc(void far32 *pMem, ULONG ulSize, char *pFile, int iLine, BOOL fCheckHeap)
{
	ALLOC far32 *pAllocOld, far32 *pAllocNew;
	char sz[256];

	if (!pMem)
	{
		void far32 *pv;

		sprintf(sz, "MEM_farrealloc(%p, %lu)", pMem, ulSize);
		OpenAllocPrintf(sz, pFile, iLine);

		if (ulSize == 0)
			pv = NULL;
		else
			pv = __farmalloc(ulSize, pFile, iLine, fCheckHeap);
		CloseAllocPrintf(TRUE, pv);
		return pv;
	}

	pAllocOld = (ALLOC *) (((ALLOC huge32 *) pMem) - 1);	/* get the struct */

	sprintf(sz, "MEM_farrealloc(/* %s %d: */ %p, %lu)",
				pAllocOld->a.pFileName, pAllocOld->a.iFileLine, pMem, ulSize);
	OpenAllocPrintf(sz, pFile, iLine);

	/* Is structure still valid? */

	if (!FCheckSumOK(pAllocOld))
	{
		_AllocPrintf(pAllocOld, "Passed in");
		if (fCheckHeap)
			FMaybeCheckHeap();
		CloseAllocPrintf(TRUE, NULL);
		return NULL;
	}

//	/* Fill out values for __CheckHeap() and _AllocPrintf(). */

//	pAllocOld->a.pFileName = pFile;
//	pAllocOld->a.iFileLine = iLine;

	/* See if the heap is still okay. */

	if (fCheckHeap)
	{
		if (!FMaybeCheckHeap())
		{
			CloseAllocPrintf(TRUE, NULL);
			return NULL;
		}
	}

	/* Get the new memory block of <ulSize> bytes. */

	if (ulSize == 0)
	{
		pAllocNew = NULL;
	}
	else
	{
		if ((pAllocNew = __farmalloc(ulSize, pFile, iLine, FALSE)) != NULL)
		{
			/* copy contents across (no need to zero extra, __farmalloc() did it). */

			MEM_ulmemcpy(pAllocNew, pMem, min(ulSize, pAllocOld->a.ulSize));
		}
	}
	__farfree(pMem, pFile, iLine, FALSE);	// don't check heap again!

	CloseAllocPrintf(TRUE, pAllocNew);
	return (void *) pAllocNew;
} /* __farrealloc */
#endif	/* __MSDOS16__ */

