/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * STRINGS.c
 *
 * PROGRAMMER : Juan M. Alvarado & Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 11/01/89
 *   MODIFIED : 02/20/95
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w-----
 *	 E C H I D N A
 *	--------------
 *
 * DESCRIPTION
 *		Routines for dealing with strings.
 *
 * HISTORY
 *		06/06/91 (RGM) - Let SAS/Lattice get dupstr() from this file.
 *		10/06/94 (dcc) - updated to use <echidna/alloc.h>
 *		10/10/94 (dcc) - make DEBUG dupstr() call our routine (call MEM_malloc())
 *
*/

#include <echidna\platform.h>
#include "switches.h"

#include <echidna/alloc.h>
#include <echidna\strings.h>

#include <echidna/etypes.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_stristr                                               */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * stristr
 *
 * SYNOPSIS
 *		char *stristr (char *data, char *key)
 *
 * PURPOSE
 *		Find key inside of data.  Similar to the ANSI routine strstr()
 *		but this version is case insensitive.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		address of first character of key inside data or
 *		NULL if not found.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		strstr
 *	
*/
char *stristr (char *data, char *key)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "stristr";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	if (*key)
	{
		while (*data)
		{
			char	*d;
			char	*k;

			d = data;
			k = key;
			while (toupper(*d) == toupper(*k))
			{
				d++;
				k++;
				if (!*k)
				{
					return data;
				}
			}
			data++;
		}
	}
	return NULL;

} /* stristr */


/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_lchinstr                                              */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * lchinstr
 *
 * SYNOPSIS
 *		int lchinstr (char ch, char *str)
 *
 * USAGE
 *		FirstDotPosition = lchinstr ('.', filenamestr);
 *
 * PURPOSE
 *		To find the position of the first occurence from the left of the
 *		character 'ch' in string 'str'.
 *
 * INPUTS
 *		ch		: character to find.
 *		str		: string in which to find character. (may be NULL)
 *
 * RESULTS
 *		If character is in string then returns positon (0..strlen (str)) of
 *		leftmost occurence of character in string. Else returns < 0.
 *
 * BUGS
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
int lchinstr (
	char	 ch,
	char	*str
)
{
	int position = -1;
	int i;

	if (!str)
		return -1;
	for (i = 0; *str && (position < 0); i++, str++){
		if (ch == *str)
			position = i;
 	}
	return position;
} /* lchinstr */

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_rchinstr                                              */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * rchinstr
 *
 * SYNOPSIS
 *		int rchinstr (char ch, char *str)
 *
 * USAGE
 *		LastSlashPosition = rchinstr ('/', filenamestr);
 *
 * PURPOSE
 *		To find the position of the Last occurence from the left of the
 *		character 'ch' in string 'str'.
 *
 * INPUTS
 *		ch		: character to find.
 *		str		: string in which to find character. (may be NULL)
 *
 * RESULTS
 *		If character is in string then returns positon (0..strlen (str)) of
 *		rightmost occurence of character in string. Else returns < 0.
 *
 * BUGS
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
int rchinstr (
	char	 ch,
	char	*str
)
{
	int position = -1;
	int i;

	if (!str)
		return -1;
	for (i = 0; *str; i++, str++){
		if (ch == *str)
			position = i;
 	}
	return position;
} /* rchinstr */

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_rstrchinstr                                           */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * rstrchinstr
 *
 * SYNOPSIS
 *		int rstrchinstr (char *chset, char *str)
 *
 * USAGE
 *		LastPathDelimPosition = rstrchinstr ("\/", path);
 *
 * PURPOSE
 *		To find the position of the last occurence in string 'str' of
 *		any character from string 'chset' except NULL character.
 *
 * INPUTS
 *		chset	: string of characters to find.
 *		str		: string in which to find characters. (may be NULL)
 *
 * RESULTS
 *		If character from 'chset' is in string then returns positon
 *		(0..strlen (str)) of rightmost occurence of characters in string.
 *		Else returns < 0.
 *
 * BUGS
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
int rstrchinstr (
	char	*chset,
	char	*str
)
{
	int position = -1;
	int i;

	if (!str || !chset)
		return -1;
	for (i = 0; *str; i++, str++){
		if (lchinstr (*str,chset) >= 0)
			position = i;
 	}
	return position;
} /* rstrchinstr */

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_dupstr                                                */
/*------------------------------------------------------------------------*/

#if (AZTEC_C || LATTICE || DEBUG)
char *dupstr (char *s)
{
	char	*d;

	d = MEM_malloc (strlen (s) + 1);
	if (d) {
		strcpy (d, s);
	}
	return (d);
}
#endif

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_stricmp                                               */
/*------------------------------------------------------------------------*/

#if (AZTEC_C || __MACSC__ || __EGGS_SGI__)
int stricmp (const char *s1, const char *s2)
{
	while (*s1 && *s2) {
		if (toupper (*s1) < toupper (*s2)) return -1;
		if (toupper (*s1) > toupper (*s2)) return  1;
		s1++;
		s2++;
	}

	if (*s1) {
		return  1;
	} else if (*s2) {
		return -1;
	}

	return 0;
}
#endif

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_strnicmp                                              */
/*------------------------------------------------------------------------*/

#if (AZTEC_C || __MACSC__ || __EGGS_SGI__)
int strnicmp (const char *s1, const char *s2, size_t len)
{
	while (*s1 && *s2 && len) {
		if (toupper (*s1) < toupper (*s2)) return -1;
		if (toupper (*s1) > toupper (*s2)) return  1;
		s1++;
		s2++;
		len--;
	}

	return 0;
}
#endif

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_strupr                                                */
/*------------------------------------------------------------------------*/

#if (AZTEC_C || __EGGS_SGI__)
char *strupr (char *s)
{
	register char *h = s;

	while (*h) {
		*h = toupper (*h);
		h++;
	}
	return s;
}
#endif

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_strlwr                                                */
/*------------------------------------------------------------------------*/

#if (AZTEC_C || __EGGS_SGI__)
char *strlwr (char *s)
{
	register char *h = s;

	while (*h) {
		*h = tolower (*h);
		h++;
	}
	return s;
}
#endif

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_TrimWhiteSpace                                        */
/*------------------------------------------------------------------------*/

/**************************************************************************
 *
 * TrimWhiteSpace 
 *
 * SYNOPSIS
 *		char *TrimWhiteSpace (char *str)
 *
 * PURPOSE
 *		Remove both leading and trailing white space from string. 
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * BUGS
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
char *TrimWhiteSpace (char *str)
{
	size_t	 len;

	len = strlen (str);
	while (len && isspace (str[len - 1])) {
		len--;
	}
	str[len] = '\0';
	while (isspace (*str)) {
		str++;
	}
	return str;

} /* TrimWhiteSpace  */

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_UnEscStr                                              */
/*------------------------------------------------------------------------*/

/**************************************************************************
 *
 * UnEscStr
 *
 * SYNOPSIS
 *		char *UnEscStr (char *str)
 *
 * PURPOSE
 *		convert esc string to un-escaped string IN PLACE using
 *		'C' style escape sequences (ie '\n' '\t' '\o123'...);
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * BUGS
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
char *UnEscStr (char *orig)
{
	int				 c;
	register char	*s    = orig;
	register char	*dest = orig;
	char			 digit[10];
	short			 digits = 0;
	short			 esc = FALSE;
	short			 hex = FALSE;
	short			 oct = FALSE;
	int				 valid;

	while (*s) {
		if (oct) {
			valid = (*s >= '0' && *s <= '7');
			if (valid || digits < 3) {
				if (valid) {
					digit[digits++] = *s++;
				}
				if (!(*s >= '0' && *s <= '7') || digits == 3) {
					oct = FALSE;
					digit[digits] = '\0';
					sscanf (digit, "%o", &c);
					*dest++ = (char) c;
				} else {
					continue;
				}
			}
		} else if (hex) {
			valid = isxdigit (*s);
			if (valid || digits < 2) {
				if (valid) {
					digit[digits++] = *s++;
				}
				if (!isxdigit(*s) || digits == 2) {
					hex = FALSE;
					digit[digits] = '\0';
					sscanf (digit, "%x", &c);
					*dest++ = (char) c;
				} else {
					continue;
				}
			}
		}
		if (!esc && *s == '\\') {
			esc = TRUE;
			s++;
		} else if (esc) {
			esc = FALSE;
			  c = -1;
			switch (*s) {
			case 'b': c = '\b'; break;
			case 'f': c = '\f'; break;
			case 'n': c = '\n'; break;
			case 'r': c = '\r'; break;
			case 't': c = '\t'; break;
			case 'v': c = '\v'; break;
			case 'x': hex = TRUE; digits = 0; break;
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
					  oct = TRUE; digits = 1; digit[0] = *s; break;
			default:  c = *s; break;
			}
			if (c >= 0) {
				*dest++ = (char) c;
			}
			s++;
		} else {
			*dest++ = *s++;
		}
	}
	*dest = '\0';

	return orig;

} /* UnEscStr */

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_itoa                                                  */
/*------------------------------------------------------------------------*/

#if (AZTEC_C)
char *itoa (int value, char *string, int radix)
{
	char	*s = string;
	UWORD	 d;
	UWORD	 v;
	UWORD	 w;

	if (radix == 10 && value < 0) {
		*s++ = '-';
		v    = -value;
	} else {
		v    = value;
	}

	d = 1;

	do {
		d = d * radix;
	} while (d <= v);

	d = d / radix;

	do {
		w    = v / d;
		*s++ = w + '0';
		v    = v - w * d;
		d    = d / radix;
	} while (d);

	*s++ = '\0';

	return string;
}
#endif

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_ltoa                                                  */
/*------------------------------------------------------------------------*/

#if (AZTEC_C || __EGGS_SGI__)
char *ltoa (long value, char *string, int radix)
{
	char	*s = string;
	ULONG	 d;
	ULONG	 v;
	ULONG	 w;

	if (radix == 10 && value < 0) {
		*s++ = '-';
		v    = -value;
	} else {
		v    = value;
	}

	d = 1;

	do {
		d = d * radix;
	} while (d <= v);

	d = d / radix;

	do {
		w    = v / d;
		*s++ = w + '0';
		v    = v - w * d;
		d    = d / radix;
	} while (d);

	*s++ = '\0';

	return string;
}
#endif


/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_StripComment                                          */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * StripComment
 *
 * SYNOPSIS
 *		void StripComment (char	*line)
 *
 * PURPOSE
 *		Remove ';' type comment from a string.
 *		Places a '\0' at the last non-quoted ';' in the string.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void StripComment (char	*s)
{
	short	 q1 = 0;
	short	 q2 = 0;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "StripComment";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	/*
	** Remove Comment.
	*/
	while (*s) {
		if (*s == '\'') {
			if (q1) { q1--; } else { q1++; }
		} else if (*s == '\"') {
			if (q2) { q2--; } else { q2++; }
		} else if ((*s == ';') && (!q1) && (!q2)) {
			*s = '\0';
			break;
		}
		s++;
	}
} /* StripComment */

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_repstr                                                */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * repstr
 *
 * SYNOPSIS
 *		short repstr (char **dest, char *src)
 *
 * PURPOSE
 *		Replace String pointed to by the pointer in dest with the
 *		string pointed to by src.
 *	
 *		repstr first checks to see the strings are different.  If they
 *		are different then repstr calls dupstr on the string pointed to
 *		by src.  If the call to dupstr is successful then the string
 *		pointed to the the pointer in dest is freed using freestr
 *		and the dest pointer is pointed to the new string.
 *
 * INPUT
 *		dest	: pointer to pointer to NULL terminated string.
 *		src		: pointer to NULL terminated string.
 *
 * EXAMPLE
 *		#include <echidna/platform.h>
 *		#include "switches.h"
 *	
 *		#include <echidna/strings.h>
 *		#include <stdio.h>
 *	
 *		void main (void) {
 *			char	*s;
 *	
 *			s = dupstr ("gregg");
 *			if (repstr (&s, "richard")) {
 *				printf ("s = '%s'\n", s);
 *			}
 *			
 *		}
 *
 * RETURN VALUE
 *		TRUE = success
 *		FALSE = failure
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		dupstr, freestr
 *	
*/
short repstr (char **dest, char *src)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "repstr";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		char	*s;

		if (strcmp (*dest, src)) {
			s = dupstr (src);
			if (!s) {
				return FALSE;
			}
			freestr (*dest);
			*dest = s;
		}
		return TRUE;
	}
} /* repstr */

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_fstrpcmp                                              */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * fstrpcmp
 *
 * SYNOPSIS
 *		short	fstrpcmp (
 *			char *fpat
 *			char *fstr
 *		)
 *
 * PURPOSE
 *		To compare two file name strings, the first of which may contain
 *		wildcards:
 *			*	Match all characters up to period (.) or end of string.
 *			?	Match any one character or nothing if at period (.) or
 *				end of string.
 *			.	Match period (.) or end of string.
 *		Using this routine will give you the same matches you get when
 *		using the DOS dir command.
 *
 * INPUT
 *		fpat:	filename string (non-NULL) that may contain wildcards.
 *		fstr:	filename string (non-NULL).
 *		
 * EFFECTS
 *		None.
 *
 * RETURN VALUE
 *		Returns 0 if matched. Returns non-zero if no match.
 *
 * HISTORY
 *
 *		Note: Does the '.' really need to be searched for?  Because
 *		*.* could mean search for anthing that starts with anything
 *		has a '.' and ends with anything which would match MS-DOS files
 *		but it would also allow searches of '*A*' and such.  Of course
 *		then '*' would match any file but I think it already does that
 *		because to list files with no extension you need '*.' at least
 *		for the MS-DOS dir command
 *
 * SEE ALSO
 *
*/
short	fstrpcmp (
	char *fpat,
	char *fstr
)
{
	short	 str_spanned = 0;
	char	*periodptr;

	for (;;) {
		switch (*fpat) {
		case '*':
			if ((periodptr = strchr (fstr, '.')) != 0) 
				fstr = periodptr + 1;
			else
				fstr = "\0";
			if ((periodptr = strchr (fpat, '.')) != 0)
				fpat = periodptr + 1;
			else
				fpat = "\0";
			continue;
		case '.':
			if (*fstr == '.')
				break;
			if (!*fstr) {
				++fpat;
				continue;
			} else
				return 2;
		case '?':
			if (*fstr == '.' || !*fstr) {
				++fpat;
				continue;
			}
			break;
		case '\0':
			return (short)*fstr;
		default:
			if (toupper(*fpat) != toupper(*fstr)) {
				return 3;
			}
		}
		++fpat;
		++fstr;
		++str_spanned;
	}

} /* fstrpcmp */

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_memcpysrc                                             */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * memcpysrc
 *
 * SYNOPSIS
 *		void *memcpysrc (void *dst, void *src, size_t len)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		(void *)src + len
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void *memcpysrc (void *dst, void *src, size_t len)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "memcpysrc";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	memcpy (dst, src, len);
	return (void *)(((char *)src) + len);

} /* memcpysrc */

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_memcpydst                                             */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * memcpydst
 *
 * SYNOPSIS
 *		void *memcpydst (void *dst, void *src, size_t len)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		(void *)dst + len
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void *memcpydst (void *dst, void *src, size_t len)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "memcpydst";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	memcpy (dst, src, len);
	return (void *)(((char *)dst) + len);

} /* memcpydst */

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_eggs_atos                                             */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * eggs_atos
 *
 * SYNOPSIS
 *		short eggs_atos (char *s)
 *
 * PURPOSE
 *		returns a value of a string as a short (just like the ANSI
 *		function atoi() but this one handles number that start with
 *		"$", "0x" or end with "h" as hex
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *		02/20/95 (GAT) - Add support for negative hex numbers.
 *
 * SEE ALSO
 *
*/
short eggs_atos (char *s)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "eggs_atos";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

{
	char	*d;
	short	 hex = 0;
	short	 neg = 0;
	size_t len;

	if (!s) return 0;

	while (*s && isspace (*s))
	{
		s++;
	}
	d = s;
	len = strlen (d);

	while (*s && (isxdigit(*s) || (s == d && *s == '-')))
	{
		s++;
	}

	if (len >= 2 && d[0] == '0' && tolower(d[1]) == 'x')
	{
		hex = 3;
	}
	else if (len >= 3 && d[0] == '-' && d[1] == '0' && tolower(d[2]) == 'x')
	{
		hex = 4;
	}
	else if (d[0] == '$')
	{
		hex = 2;
	}
	else if (len >= 2 && d[0] == '-' && d[1] == '$')
	{
	hex = 3;
	}
	else if (tolower(*s) == 'h')
	{
		hex = 1;
	}

	neg = (d[0] == '-');

	if (hex)
	{
		UWORD	v = 0;

		d += (hex - 1);
		while (isxdigit (*d))
		{
			v = (UWORD) (v * 16 + ((*d <= '9') ? (*d - '0') : (toupper (*d) - 'A' + 10)));
			d++;
		}
		return (neg) ? (0-v) : (v);
	}
	else
	{
		return (short) (atoi (d));
	}
}
} /* eggs_atos */

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_eggs_atol                                             */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * eggs_atol
 *
 * SYNOPSIS
 *		long eggs_atol (char *s)
 *
 * PURPOSE
 *		returns a value of a string as a long (just like the ANSI
 *		function atol() but this one handles number that start with
 *		"$", "0x" or end with "h" as hex
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *		02/20/95 (GAT) - Add support for negative hex numbers.
 *
 * SEE ALSO
 *
*/
long eggs_atol (char *s)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "eggs_atol";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

{
	char	*d;
	short	 hex = 0;
	short	 neg = 0;
	size_t len;

	if (!s) return 0L;

	while (*s && isspace (*s))
	{
		s++;
	}
	d = s;
	len = strlen (d);

	while (*s && (isxdigit(*s) || (s == d && *s == '-')))
	{
		s++;
	}

	if (len >= 2 && d[0] == '0' && tolower(d[1]) == 'x')
	{
		hex = 3;
	}
	else if (len >= 3 && d[0] == '-' && d[1] == '0' && tolower(d[2]) == 'x')
	{
	hex = 4;
	}
	else if (d[0] == '$')
	{
		hex = 2;
	}
	else if (len >= 2 && d[0] == '-' && d[1] == '$')
	{
	hex = 3;
	}
	else if (tolower(*s) == 'h')
	{
		hex = 1;
	}

	neg = (d[0] == '-');

	if (hex)
	{
		ULONG	v = 0;

		d += (hex - 1);
		while (isxdigit (*d))
		{
			v = v * 16 + ((*d <= '9') ? (*d - '0') : (toupper (*d) - 'A' + 10));
			d++;
		}
		return (neg) ? (0-v) : (v);
	}
	else
	{
		return atoi (d);
	}
}
} /* eggs_atol */


/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_pStrCopy                                              */
/*------------------------------------------------------------------------*/

void pStrCopy(unsigned char *src,unsigned char *dest)
{
	memmove (dest,src,(size_t) *src + 1);
}

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_pStrAppend                                            */
/*------------------------------------------------------------------------*/

void pStrAppend(unsigned char * src,unsigned char * dest)
{
	unsigned short newlen;
	newlen = (unsigned short) (*dest + *src);
	memmove (dest + *dest + 1,src + 1,(int) *src);
	*dest = newlen;
}

/*------------------------------------------------------------------------*/
/**# MODULE:STRINGS_pCharAppend                                           */
/*------------------------------------------------------------------------*/

void pCharAppend( unsigned char p1,unsigned char *p2)
{
	*p2 += (unsigned char) 1;
	*(p2 + *p2) = p1;
}

