/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * ATIMER.C
 *
 * PROGRAMMER : R.G. Marquez 
 *    VERSION : 00.000
 *    CREATED : 06/23/91 
 *   MODIFIED : 07/27/94
 *       TABS : 04 09
 *
 *	     \|///-_
 *	     \oO///_
 *	-----w/-w-----
 *	 E C H i D N A
 *	--------------
 *
 * DESCRIPTION
 *		Amiga specific timer library for EGGS.
 *
 * HISTORY
 *	06/23/91 Sunday (RGM) - Created. Implementation note :
 *		A 'normal' routine will never interrupt an interrupt.  What this
 *		means is that the 'normal' list manipulation routines just have to
 *		set the atel.Lock 'semaphore', do their thing, and clear atel.Lock.
 *		The timer.lib interrupt handler routines must check atel.Lock;
 *		if it's set just exit, if not, go ahaead and use the structure.
 *		NOTE : Interrupt routines must not CHANGE the semaphored data
 *		structure for this to work.
 *	06/26/91 Wednesday (RGM) - Completed, tested for bugs and accuracy.
 *		Seems to work great.  NOTE : A subtle bug can arise if the user
 *		supplied interrupt routines aren't defined as "InterruptRoutineType".
 *		If the user function is of type InterruptRoutineType, the
 *		Amiga "near data pointer" is re-loaded on entry. . .if any 'near'
 *		data is accessed from the user interrupt but the 'near data pointer'
 *		isn't pointing to the near data area, it can trash random memory
 *		locations.  SO BE AWARE : DECLARE ALL USER ROUTINES AS :
 *			InterruptRoutineType <routine name> (void);
 *		We should also mabye make OpenTimer() return TRUE or FALSE (???).
*/

#include <echidna/platform.h>
#include "switches.h"

#define TEST_ATIMER		1
#define REAL_TIMERS		1
#define OK_CIAB_TIMER	1
#define OK_VBLANK_TIMER	1

#include <exec/types.h>

#include <exec/interrupts.h>
#include <hardware/intbits.h>
#include <hardware/cia.h>
#include <resources/cia.h>

#include <echidna/timer.h>

#include "atimer.h"
#include "atimlist.h"

/**************************** C O N S T A N T S ***************************/

/* this is the MAXIMUM rate that we want non-VBLANK interrupts to run at */
#define MAX_CYCLES_PER_SECOND	(420)

/* this is the maximum number of events we can have running at one time */
#define MAX_ATIMER_EVENTS		(64)

/* these are the rates (Hz) that the hardware timer runs at */
#define PAL_CIA_CYCLES_PER_SECOND	(709379)
#define NTSC_CIA_CYCLES_PER_SECOND	(715909)

/* this is the value that gets placed into hardware to get our MAXIMUM rate */
#define CIA_TIMER_INTERVAL \
			(NTSC_CIA_CYCLES_PER_SECOND / MAX_CYCLES_PER_SECOND)

#define CIA_TIMER_PRIORITY		(127)

/******************************** T Y P E S *******************************/


/****************************** E X T E R N S *****************************/

#if LATTICE

#include <proto/exec.h>
extern ULONG		AbleICR (ULONG);
extern void __far	*AddICRVector (ULONG, void __far *);
extern void			RemICRVector (ULONG, void __far *);
extern ULONG		SetICR (ULONG);
#pragma libcall CIAResource AbleICR 12 001
#pragma libcall CIAResource AddICRVector 06 9002
#pragma libcall CIAResource RemICRVector 0C 9002
#pragma libcall CIAResource SetICR 18 001

#else

#error Make sure that you can access these routines with your compiler:
#error     AbleICR(), AddICRVector(), RemICRVector(), SetICR().
#error You may also need to "#define __far far" or "#define far".
#error Also, make sure that do_atimer_vblank() exits the way VBLANKs
#error should exit (A0 = CUSTOM, D0 = 0). . .ATIMVB.ASM is a VBLANK
#error shell routine (called _LCVB_TimerLib) that does this for
#error do_atimer_vblank() under SAS/Lattice C.

#endif

extern struct CIA __far ciab;	/* CIA-B hardware structure */

#if LATTICE
extern void __far	LCVB_TimerLib (void);
#endif

extern void __far do_atimer_vblank (void);
extern void __far do_atimer_timer (void);

/****************************** G L O B A L S *****************************/


/******************************* L O C A L S ******************************/

static	struct Library		*CIAResource			= NULL;
static	struct Interrupt	Atimer_CIABInterrupt	= {{NULL,},NULL,NULL};
static	struct Interrupt	Atimer_VERTBinterrupt	= {{NULL,},NULL,NULL};
static	void 				*OldCIAInterrupt		= (void *) -1;


static	ATimer_EventList __far	Free_Events;
static	ATimer_EventList __far	Active_Events;
static	ATimer_EventList __far	Vblank_Events;

static	ATimer_EventNode __far	Event_Store[MAX_ATIMER_EVENTS];

static	BOOL __far				Atimer_Open			= FALSE;

/******************************* M A C R O S ******************************/


/***************************** R O U T I N E S ****************************/


/** LOCAL ROUTINES **/

void __far do_atimer_vblank (void)
{
	if (ate_IsNotLocked (&Vblank_Events)) {
		if (! ate_IsEmpty (&Vblank_Events))	{

			ATimer_EventNode __far	*current_event;

			current_event	=  ate_Head (&Vblank_Events);

			while (! ate_IsTail (current_event)) {

				/* do what we need to do */

				((InterruptRoutineType (*)(void))(current_event->ate_DataPtr)) ();

				current_event = ate_Next(current_event);
			}
		}
	}
	return;
}

void __far do_atimer_timer (void)
{
	if (ate_IsNotLocked (&Active_Events)) {
		if (! ate_IsEmpty (&Active_Events))	{

			ATimer_EventNode __far	*current_event;

			current_event	=  ate_Head (&Active_Events);

			while (! ate_IsTail (current_event)) {

				if (current_event->ate_WaitPeriod) {

					current_event->ate_IdleTimer -= 1;

					if (! current_event->ate_IdleTimer) {

						current_event->ate_IdleTimer =
										current_event->ate_WaitPeriod;

						/* we do the real stuff here */

						if (current_event->ate_Type	== ATIME_INTERRUPT) {
							((InterruptRoutineType (*)(void))
										(current_event->ate_DataPtr)) ();
						}
						else {
							short	delta;

							delta	=
								(current_event->ate_Direction >= 1) ? 1 : -1;

							if (current_event->ate_DataSize >= 4) {
								*((ULONG __far *)(current_event->ate_DataPtr)) += delta;
							}
							else if (current_event->ate_DataSize >= 2) {
								*((UWORD __far *)(current_event->ate_DataPtr)) += delta;
							}
							else if (current_event->ate_DataSize){
								*((UBYTE __far *)(current_event->ate_DataPtr)) += delta;
							}
						}
					}
				}

				current_event = ate_Next(current_event);
			}
		}
	}
	return;
}

/** GLOBAL ROUTINES **/


/************************/

void far OpenTimers (void)
{
	if (! Atimer_Open) {

		/* initialize our data */

		/* initialize the lists (to empty) */

		ate_InitList (&Free_Events);

		ate_Lock (&Active_Events);
		ate_InitList (&Active_Events);
		ate_Unlock (&Active_Events);

		ate_Lock (&Vblank_Events);
		ate_InitList (&Vblank_Events);
		ate_Unlock (&Vblank_Events);

		/* initialize the event nodes and add them to the free list */

		{
			short	i;

			for (i = 0; i < MAX_ATIMER_EVENTS; i++) {
				Event_Store[i].ate_Handle	= i;
				ate_AddTail (&Free_Events, &Event_Store[i]);
			}
		}

		/*********************************************/
		/* Get the Amiga system resources we'll need */

#if REAL_TIMERS

		/************************************************/
		/* first the CIA-B timers, (used for the timers */
		/* and the non-VBLANK interrupts).  We want to  */
		/* get the CIA-B 'B' timer                      */

		/* open the CIA-B resource */

#if OK_CIAB_TIMER
		if (CIAResource = OpenResource (CIABNAME)) {

			/* set the rate */

			ciab.ciatblo	= ((UBYTE) (CIA_TIMER_INTERVAL & 0xFF));
			ciab.ciatbhi	= ((UBYTE) (CIA_TIMER_INTERVAL >> 8));

			/* initialize the interrupt structure */

			Atimer_CIABInterrupt.is_Node.ln_Type	= NT_INTERRUPT;
			Atimer_CIABInterrupt.is_Node.ln_Name	= "timer.lib-CIAB";
			Atimer_CIABInterrupt.is_Node.ln_Pri		= CIA_TIMER_PRIORITY;

			Atimer_CIABInterrupt.is_Data	= NULL;
			Atimer_CIABInterrupt.is_Code	= (void __far *) do_atimer_timer;

			/* install the interrupt code */

			OldCIAInterrupt	=
				AddICRVector (CIAICRB_TB,(void __far *)&Atimer_CIABInterrupt);

			/* this is where we actually see if someone else */
			/* is using the CIAB 'B' timer                   */

			if (OldCIAInterrupt) {
/**/			goto ABORT;
			}

			/* start the timer */

			ciab.ciacrb	&= ~CIACRBF_RUNMODE; /* reload on underflow */
			ciab.ciacrb	|= CIACRBF_LOAD | CIACRBF_START; /* load & start */

			SetICR (CIAICRF_TB);
			AbleICR (CIAICRF_SETCLR | CIAICRF_TB);
		}
#endif

		/****************************/
		/* Next, the vblank routine */

#if OK_VBLANK_TIMER
		Atimer_VERTBinterrupt.is_Node.ln_Type	= NT_INTERRUPT;
		Atimer_VERTBinterrupt.is_Node.ln_Name	= "timer.lib-VERTB";
		Atimer_VERTBinterrupt.is_Node.ln_Pri	= 0;

		Atimer_VERTBinterrupt.is_Data	= NULL;
#if LATTICE
		Atimer_VERTBinterrupt.is_Code	= (void __far *) LCVB_TimerLib;
#else
		Atimer_VERTBinterrupt.is_Code	= (void __far *) do_atimer_vblank;
#endif

		AddIntServer ((long) INTB_VERTB, &Atimer_VERTBinterrupt);
#endif

#endif

		Atimer_Open	= TRUE;
	}
	return;
/*------------------------------------------------------------------*/
ABORT:

	CloseTimers ();

	return;
}


void far CloseTimers (void)
{
	/* Stop the interrupts and give back the system resoiurces */

#if REAL_TIMERS

	if (Atimer_VERTBinterrupt.is_Code) {
		RemIntServer((ULONG) INTB_VERTB, (void __far *) &Atimer_VERTBinterrupt);
		Atimer_VERTBinterrupt.is_Code	= NULL;
	}

	if (! OldCIAInterrupt) {
		AbleICR (CIAICRF_TB);
		ciab.ciacrb	&= ~CIACRBF_START;
		RemICRVector (CIAICRB_TB, (void __far *) &Atimer_CIABInterrupt);
		OldCIAInterrupt					= (void *) -1;
	}
#endif

	Atimer_Open	= FALSE;

	return;
}


/* ChangeTimerRate = ChangeInterruptRate */

BOOL far ChangeInterruptRate (
	WORD handle,
	WORD rate
)
{
	BOOL	status = FALSE;

	if (handle && (handle < MAX_ATIMER_EVENTS)) {
		ATimer_EventNode __far	*event;

		event = &Event_Store[handle - 1];

		if ((event->ate_Type == ATIME_TIMER) || (event->ate_Type == ATIME_INTERRUPT)) {

			WORD	period;

			if (rate) {
				period	= MAX_CYCLES_PER_SECOND / rate;
				if (! period) {
					period	= 1;
				}
			}
			else {
				period = 0;
			}

			ate_Lock (&Active_Events);
			event->ate_WaitPeriod	= event->ate_IdleTimer = period;
			ate_Unlock (&Active_Events);

			status = TRUE;
		}
	}
	return (status);
}


BOOL far ResetTimer(
	WORD	handle,
	ULONG	value
)
{
	BOOL	status = FALSE;

	if (handle && (handle < MAX_ATIMER_EVENTS)) {
		ATimer_EventNode __far	*event;

		event = &Event_Store[handle - 1];

		if (event->ate_Type == ATIME_TIMER) {

			ate_Lock (&Active_Events);

			if (event->ate_DataSize >= 4) {
				*((ULONG __far *)(event->ate_DataPtr)) = value;
			}
			else if (event->ate_DataSize >= 2) {
				*((UWORD __far *)(event->ate_DataPtr)) = value;
			}
			else if (event->ate_DataSize){
				*((UBYTE __far *)(event->ate_DataPtr)) = value;
			}

			event->ate_IdleTimer = event->ate_WaitPeriod;
			ate_Unlock (&Active_Events);

			status = TRUE;
		}
	}
	return (status);
}


/************************/

WORD far InstallInterrupt (
	InterruptRoutineType	(*routine)(void),
	WORD					rate
)
{
	WORD	handle	= 0;


	if (! ate_IsEmpty(&Free_Events)) {

		ATimer_EventNode __far	*new_event;

		/* get a free event node */

		new_event	= ate_RemHead (&Free_Events);

		/* initialize it for timer processing */

		/* interrupt routines don't need these, we'll set them to zero */
		new_event->ate_DataSize		= 0;
		new_event->ate_Direction	= 0;

		/* stuff needed for interrupt routines */
		{
			WORD	period;

			if (rate) {
				period	= MAX_CYCLES_PER_SECOND / rate;
				if (! period) {
					period	= 1;
				}
			}
			else {
				period = 0;
			}

			new_event->ate_WaitPeriod	= new_event->ate_IdleTimer = period;
		}
		new_event->ate_Type			= ATIME_INTERRUPT;
		new_event->ate_DataPtr		= (void __far *) routine;

		handle	= new_event->ate_Handle + 1;

		ate_Lock (&Active_Events);
		ate_AddTail(&Active_Events, new_event);
		ate_Unlock (&Active_Events);
	}

	return (handle);
}



WORD far InstallTimer (
	void far	*timer,
	WORD		rate,
	WORD		size,
	WORD		direction
)
{
	WORD	handle	= 0;

	/* if we can, get a free event node */

	if (! ate_IsEmpty(&Free_Events)) {

		ATimer_EventNode __far	*new_event;

		/* get a free event node */

		new_event	= ate_RemHead (&Free_Events);

		/* initialize it for timer processing */
		{
			WORD	period;

			if (rate) {
				period	= MAX_CYCLES_PER_SECOND / rate;
				if (! period) {
					period	= 1;
				}
			}
			else {
				period = 0;
			}

			new_event->ate_WaitPeriod	= new_event->ate_IdleTimer = period;
		}
		new_event->ate_Direction	= direction;
		new_event->ate_DataSize		= size;

		new_event->ate_Type			= ATIME_TIMER;
		new_event->ate_DataPtr		= (void __far *) timer;

		handle	= new_event->ate_Handle + 1;

		ate_Lock (&Active_Events);
		ate_AddTail(&Active_Events, new_event);
		ate_Unlock (&Active_Events);
	}

	return (handle);
}


WORD far InstallVBlank (InterruptRoutineType (*routine)(void))
{
	WORD	handle	= 0;

	/* if we can, get a free event node */

	if (! ate_IsEmpty(&Free_Events)) {

		ATimer_EventNode __far	*new_event;

		/* get a free event node */

		new_event	= ate_RemHead (&Free_Events);

		/* initialize it for vblank processing */

		/* vblank routines don't need these, we'll set them to zero */
		new_event->ate_IdleTimer	= 0;
		new_event->ate_WaitPeriod	= 0;
		new_event->ate_Direction	= 0;
		new_event->ate_DataSize		= 0;

		/* stuff needed for vblank routines */
		new_event->ate_Type		= ATIME_VBLANK;
		new_event->ate_DataPtr	= (void __far *) routine;

		handle	= new_event->ate_Handle + 1;

		ate_Lock (&Vblank_Events);
		ate_AddTail(&Vblank_Events, new_event);
		ate_Unlock (&Vblank_Events);
	}

	return (handle);
}


/* RemoveTimer = RemoveInterrupt */
BOOL far RemoveInterrupt (WORD handle)
{
	BOOL	status = FALSE;

	if (handle && (handle < MAX_ATIMER_EVENTS)) {
		ATimer_EventNode __far	*event;

		event = &Event_Store[handle - 1];

		if ((event->ate_Type == ATIME_TIMER) || (event->ate_Type == ATIME_INTERRUPT)) {

			ate_Lock (&Active_Events);
			ate_Remove (event);
			ate_Unlock (&Active_Events);

			event->ate_Type = ATIME_FREE;
			ate_AddTail (&Free_Events, event);

			status = TRUE;
		}
	}
	return (status);
}

BOOL far RemoveVBlank (WORD handle)
{
	BOOL	status = FALSE;

	if (handle && (handle < MAX_ATIMER_EVENTS)) {
		ATimer_EventNode __far	*event;

		event = &Event_Store[handle - 1];

		if (event->ate_Type == ATIME_VBLANK) {

			ate_Lock (&Vblank_Events);
			ate_Remove (event);
			ate_Unlock (&Vblank_Events);

			event->ate_Type = ATIME_FREE;
			ate_AddTail (&Free_Events, event);

			status = TRUE;
		}
	}
	return (status);
}

/************************/

#if TEST_ATIMER

#include <stdio.h>
#include <stdlib.h>


UBYTE one	= 0;
UWORD two	= 0;
ULONG four	= 0L;

ULONG	SecondsLeft	= 10L;
ULONG	VblankCount	= 0L;

InterruptRoutineType OncePerSecond (void)
{
	SecondsLeft	-= 1;

	return;
}

InterruptRoutineType VBlanker (void)
{
	VblankCount	+= 1;

	return;
}

#include <exec/execbase.h>
extern struct ExecBase	*SysBase;
UBYTE get_frequency (void)
{
	UBYTE	hertz;
	hertz	= SysBase->VBlankFrequency;
	return (hertz);
}

void main (void)
{
	printf ("Amiga timer.lib test, running VBLANK @ %dHz.\n", get_frequency());

	printf ("Will attempt a %d second wait.\n", SecondsLeft);

	OpenTimers ();

	if (! InstallInterrupt ((void __far *) &OncePerSecond, 1)) {
		printf ("Didn't install seconds timer.\n");
	}
	else {
		InstallVBlank ((void __far *) &VBlanker);

		InstallTimer ((void __far *) &four, 60, 4, 1);

		while (SecondsLeft) {
		}
	}

	CloseTimers ();

	printf ("VblankCount                 : %d.\n", VblankCount);
	printf ("Sixtieths of a second timer : %d.\n", four);
}

#endif

